import copy
import json
from pprint import pprint
from typing import Optional

import elasticsearch.exceptions as es_exc
from inverter import dc2colanderESjson, dc2esmapping
from rulez import compile_condition

from ..app import App
from .base import BaseStorage


class AggGroup(object):
    def __init__(self, key, field, type="terms", opts=None, children=None):
        self.key = key
        self.field = field
        self.type = type
        self.children = children or []
        self.parent = None
        self.is_leaf = False
        self.opts = opts or {}

    def parse(self, result, parents=None):
        parents = parents or {}
        records = result[self.key]["buckets"]
        results = []
        for r in records:
            key = r["key_as_string"]
            if self.type == "date_histogram":
                if self.opts.get("format", None) in ["yyyy", "MM", "dd"]:
                    key = int(key)
            parents[self.key] = key
            if self.is_leaf:
                values = copy.copy(parents)
                for c in self.children:
                    if c.function == "count":
                        values[c.key] = r["doc_count"]
                    else:
                        values.update(c.parse(r, parents=parents))
                results.append(values)
            else:
                values = []
                for c in self.children:
                    values += c.parse(r, parents=parents)
                results += values
        return results

    def json(self):
        res = {self.type: {"field": self.field}}

        if self.opts:
            res[self.type].update(self.opts)

        if self.children:
            for c in self.children:
                res.setdefault("aggs", {})
                res["aggs"].update(c.json())

        return {self.key: res}


class AggValue(object):
    def __init__(self, key, function, field):
        self.key = key
        self.function = function
        self.field = field
        self.parent = None

    def parse(self, result, parents=None):
        if self.function == "count":
            return {}
        parents = parents or {}
        p = copy.copy(parents)
        p[self.key] = result[self.key][self.field]
        return p

    def json(self):
        if self.function == "count":
            return {}
        return {self.key: {self.function: {"field": self.field}}}


class Aggregate(object):
    def __init__(self):
        self.groups = []
        self.values = []
        self.first_group = None
        self._finalized = False

    def add_group(self, key, field, type="terms", opts=None):
        if self._finalized:
            raise ValueError("Aggregate has been finalized")
        self.groups.append(AggGroup(key, field, type, opts))

    def add(self, key, function, field):
        if self._finalized:
            raise ValueError("Aggregate has been finalized")
        self.values.append(AggValue(key, function, field))

    def finalize(self):
        if self._finalized:
            raise ValueError("Aggregate has been finalized")

        if not self.groups:
            self._finalized = True
            return

        prev = None
        first = None
        for g in self.groups:
            if prev is None:
                prev = g
                continue
            if g not in prev.children:
                prev.children.append(g)
                g.parent = prev
            prev = g
        prev.is_leaf = True
        for v in self.values:
            if v not in prev.children:
                prev.children.append(v)
                v.parent = prev

        self._finalized = True

    def json(self):
        if not self._finalized:
            self.finalize()
        if not self.groups:
            return {}
        return self.groups[0].json()

    def parse(self, result):
        if not self._finalized:
            self.finalize()
        return self.groups[0].parse(result)


def is_text_mapping(collection, field):
    schema = collection.schema
    f = schema.__dataclass_fields__[field]
    fmt = f.metadata.get("format", None)
    if not fmt:
        return False

    if fmt.startswith("text"):
        return True
    return False


class ElasticSearchStorage(BaseStorage):

    refresh: Optional[str] = None
    auto_id = False
    use_transactions = False

    @property
    def index_name(self):
        raise NotImplementedError

    @property
    def client(self):
        return self.request.get_es_client()

    def create_index(self, collection, recreate=False):
        if self.client.indices.exists(self.index_name) and recreate == False:
            return False

        if recreate == True:
            self.client.indices.delete(self.index_name)

        settings = self.request.app.get_config(
            "morpfw.storage.elasticsearch.index_settings.%s" % self.index_name, None,
        )
        body = {}
        if settings:
            body["settings"] = settings or {
                "number_of_shards": 1,
                "number_of_replicas": 0,
            }

        mapping = dc2esmapping.convert(collection.schema)
        body.update(mapping)
        self.client.indices.create(
            index=self.index_name, body=body,
        )

        return True

    def update_index(self, collection):
        if not self.client.indices.exists(self.index_name):
            return False

        fields = list(collection.schema.__dataclass_fields__.keys())
        mappings = self.client.indices.get_field_mapping(
            fields, index=self.index_name, ignore_unavailable=True
        )

        existing_fields = list(mappings[self.index_name]["mappings"].keys())
        new_fields = [x for x in fields if x not in existing_fields]
        new_mapping = dc2esmapping.convert(
            collection.schema, include_fields=list(new_fields)
        )
        self.client.indices.put_mapping(
            index=self.index_name, body=new_mapping["mappings"],
        )
        return True

    def create(self, collection, data):
        m = self.model(self.request, collection, data)
        cschema = dc2colanderESjson.convert(
            collection.schema,
            request=collection.request,
            default_tzinfo=collection.request.timezone(),
        )
        esdata = cschema().serialize(data)
        try:
            r = self.client.index(
                index=self.index_name,
                id=m.identifier,
                body=esdata,
                refresh=self.refresh,
            )
        except es_exc.TransportError as e:
            print(e.args, e.error, e.info)
            print(data)
            raise e

        if self.auto_id:
            self.set_identifier(m.data, r["_id"])
            m.save()
        return m

    def search(self, collection, query=None, offset=None, limit=None, order_by=None):
        if limit is None:
            limit = 9999
        if query:
            q = {"query": compile_condition("elasticsearch", query)()}
        else:
            q = {"query": {"match_all": {}}}
        if limit is not None:
            q["from"] = 0
            q["size"] = limit

        params = {}
        if offset is not None:
            params["from_"] = offset
        if limit:
            params["size"] = limit
        if order_by:
            if is_text_mapping(collection, order_by[0]):
                params["sort"] = [":".join(["%s.raw" % order_by[0], order_by[1]])]
            else:
                params["sort"] = [":".join(order_by)]

        res = self.client.search(index=self.index_name, body=q, **params)

        models = []
        for o in res["hits"]["hits"]:
            data = o["_source"]
            cschema = dc2colanderESjson.convert(
                collection.schema,
                include_fields=data.keys(),
                request=collection.request,
                default_tzinfo=collection.request.timezone(),
            )
            data = cschema().deserialize(data)
            models.append(self.model(self.request, collection, data))

        return list(models)

    def aggregate(self, query=None, group=None, order_by=None, limit=None):
        if group is None:
            return []

        if query:
            q = {"query": compile_condition("elasticsearch", query)()}
        else:
            q = {"query": {"match_all": {}}}

        q["size"] = 0

        params = {}
        if order_by:
            params["sort"] = [":".join(order_by)]
        if limit is not None:
            params["size"] = limit

        aggs = {}
        sources = []
        result_converters = {}
        if group:

            for k, v in group.items():
                if isinstance(v, str):
                    sources.append({k: {"terms": {"field": v}}})

                elif isinstance(v, dict):
                    ff = v["function"]
                    f = v["field"]
                    if ff == "count":
                        aggs[k] = {"value_count": {"field": f}}
                    elif ff == "sum":
                        aggs[k] = {"sum": {"field": f}}
                    elif ff == "avg":
                        aggs[k] = {"avg": {"field": f}}
                    elif ff == "year":
                        sources.append(
                            {
                                k: {
                                    "date_histogram": {
                                        "field": f,
                                        "interval": "year",
                                        "format": "yyyy",
                                    }
                                }
                            }
                        )
                        result_converters[k] = lambda request, value: int(value)

                    elif ff == "month":
                        sources.append(
                            {
                                k: {
                                    "date_histogram": {
                                        "field": f,
                                        "interval": "month",
                                        "format": "MM",
                                    }
                                }
                            }
                        )

                        result_converters[k] = lambda request, value: int(value)

                    elif ff == "day":
                        sources.append(
                            {
                                k: {
                                    "date_histogram": {
                                        "field": f,
                                        "interval": "day",
                                        "format": "dd",
                                    }
                                }
                            }
                        )

                        result_converters[k] = lambda request, value: int(value)

                    elif ff == "interval_1m":
                        sources.append(
                            {
                                k: {
                                    "date_histogram": {
                                        "field": f,
                                        "interval": "1m",
                                        "format": "yyyy-MM-dd'T'HH:mmZ",
                                    }
                                }
                            }
                        )

                    elif ff == "interval_15m":
                        sources.append(
                            {
                                k: {
                                    "date_histogram": {
                                        "field": f,
                                        "interval": "15m",
                                        "format": "yyyy-MM-dd'T'HH:mmZ",
                                    }
                                }
                            }
                        )
                    elif ff == "interval_30m":
                        sources.append(
                            {
                                k: {
                                    "date_histogram": {
                                        "field": f,
                                        "interval": "30m",
                                        "format": "yyyy-MM-dd'T'HH:mmZ",
                                    }
                                }
                            }
                        )

                    elif ff == "interval_1h":
                        sources.append(
                            {
                                k: {
                                    "date_histogram": {
                                        "field": f,
                                        "interval": "1h",
                                        "format": "yyyy-MM-dd'T'HH:mmZ",
                                    }
                                }
                            }
                        )

                    else:
                        raise ValueError("Unknown function %s" % ff)

            if sources:
                q["aggs"] = {
                    "results": {"composite": {"sources": sources}, "aggs": aggs}
                }
                if limit is not None:
                    q["aggs"]["results"]["composite"]["size"] = limit
            else:
                q["aggs"] = aggs

        res = self.client.search(index=self.index_name, body=q, **params)

        if "aggregations" not in res:
            if len(group) == 1 and "count" in group.keys():
                return [{"count": res["hits"]["total"]["value"]}]
            raise AssertionError(
                "Unsupported query\n  Query: %s\n  Result: %s"
                % (json.dumps(q, indent=4), json.dumps(res, indent=4))
            )

        result = []
        if sources:
            data = res["aggregations"]["results"]["buckets"]
            for row in data:
                r = row["key"].copy()
                for k in aggs:
                    r[k] = row[k]["value"]
                for k, f in result_converters.items():
                    r[k] = f(self.request, r[k])
                result.append(r)
            return result

        data = res["aggregations"]
        r = {}
        for k in aggs:
            r[k] = data[k]["value"]
        return [r]

    def get(self, collection, identifier):
        try:
            res = self.client.get(
                index=self.index_name, id=identifier, refresh=self.refresh,
            )
        except es_exc.NotFoundError as e:
            return None

        data = res["_source"]
        cschema = dc2colanderESjson.convert(
            collection.schema,
            include_fields=data.keys(),
            request=collection.request,
            default_tzinfo=collection.request.timezone(),
        )
        data = cschema().deserialize(data)
        return self.model(self.request, collection, data)

    def get_by_uuid(self, collection, uuid):
        res = self.search(
            collection, {"field": "uuid", "operator": "==", "value": uuid}
        )
        if res:
            data = res[0].as_dict()
            return self.model(self.request, collection, data)
        return None

    def get_by_id(self, collection, id):
        res = self.client.get(index=self.index_name, id=id)
        data = res["_source"]
        cschema = dc2colanderESjson.convert(
            collection.schema,
            include_fields=data.keys(),
            request=collection.request,
            default_tzinfo=collection.request.timezone(),
        )
        data = cschema().deserialize(data)
        return self.model(self.request, collection, data)

    def update(self, collection, identifier, data):
        cschema = dc2colanderESjson.convert(
            collection.schema,
            include_fields=data.keys(),
            request=collection.request,
            default_tzinfo=collection.request.timezone(),
        )
        data = cschema().serialize(data)
        self.client.update(
            index=self.index_name,
            id=identifier,
            body={"doc": data},
            refresh=self.refresh,
        )

    def delete(self, identifier, model, **kwargs):
        self.client.delete(index=self.index_name, id=identifier, refresh=True)




from ...app import BaseApp
from .apikey.model import APIKeyCollection, APIKeyModel, APIKeySchema
from .apikey.path import get_apikey, get_apikey_collection
from .group.model import GroupCollection, GroupModel, GroupSchema
from .group.path import get_group, get_group_collection
from .user.model import CurrentUserModel, UserCollection, UserModel
from .user.path import get_user, get_user_collection


def hook_auth_models(cls, prefix=""):
    @cls.path(
        model=UserModel,
        path="%s/user/{username}" % prefix,
        variables=lambda obj: {"username": obj.data["username"]},
    )
    def _get_user(app, request, username):
        return get_user_collection(request).get_by_username(username)

    @cls.path(model=UserCollection, path="%s/user" % prefix)
    def _get_user_collection(app, request):
        return get_user_collection(request)

    @cls.path(model=CurrentUserModel, path="%s/self" % prefix)
    def _get_current_user(app, request):
        userid = request.identity.userid
        if not userid:
            return None
        col = get_user_collection(request)
        user = col.get_by_userid(userid)
        if not user:
            return None
        return CurrentUserModel(request, user.collection, user.data.data)

    @cls.path(model=GroupModel, path="%s/group/{identifier}" % prefix)
    def _get_group(app, request, identifier):
        return get_group_collection(request).get_by_groupname(identifier)

    @cls.path(model=GroupCollection, path="%s/group" % prefix)
    def _get_group_collection(app, request):
        return get_group_collection(request)

    @cls.path(model=APIKeyModel, path="%s/apikey/{uuid}" % prefix)
    def _get_apikey(app, request, uuid):
        return get_apikey(request, uuid)

    @cls.path(model=APIKeyCollection, path="%s/apikey" % prefix)
    def _get_apikey_collection(app, request):
        return get_apikey_collection(request)


BaseApp.hook_auth_models = classmethod(hook_auth_models)



import os
import tempfile
import typing
from dataclasses import dataclass, field
from datetime import date
from urllib.parse import quote

import pytest
import rulez
from morpfw import Schema
from morpfw.app import BaseApp
from morpfw.crud import permission as crudperm
from morpfw.crud.model import Collection, Model
from morpfw.crud.statemachine.base import StateMachine
from morpfw.crud.workflow import WorkflowCollection, WorkflowModel

from ...app import SQLApp
from ...crud.storage.sqlstorage import SQLStorage
from ...sql import Base, construct_orm_model
from ..common import get_client, make_request


class App(SQLApp):
    pass


@App.permission_rule(model=Model, permission=crudperm.All)
def allow_all_model_access(identity, context, permission):
    return True


@App.permission_rule(model=Model, permission=crudperm.StateUpdate)
def allow_all_model_state_access(identity, context, permission):
    return True


@App.permission_rule(model=Collection, permission=crudperm.All)
def allow_all_collection_access(identity, context, permission):
    return True


@dataclass
class ObjectSchema(Schema):
    pass


class ObjectCollection(Collection):
    schema = ObjectSchema


class ObjectModel(Model):
    schema = ObjectSchema


@App.path(model=ObjectCollection, path="objects")
def object_collection_factory(request):
    storage = request.app.get_storage(ObjectModel, request)
    return ObjectCollection(request, storage)


@App.path(model=ObjectModel, path="objects/{identifier}")
def object_model_factory(request, identifier):
    collection = object_collection_factory(request)
    return collection.get(identifier)


@App.typeinfo(name="tests.object", schema=ObjectSchema)
def get_object_typeinfo(request):
    return {
        "title": "Object",
        "description": "Object type",
        "schema": ObjectSchema,
        "collection": ObjectCollection,
        "collection_factory": object_collection_factory,
        "model": ObjectModel,
        "model_factory": object_model_factory,
    }


class ObjectStorage(SQLStorage):
    model = ObjectModel
    orm_model = construct_orm_model(
        schema=ObjectModel.schema, metadata=Base.metadata, name="test_object"
    )


@App.storage(model=ObjectModel)
def get_object_storage(model, request, blobstorage):
    return ObjectStorage(request, blobstorage)


def test_metalink(pgsql_db):
    config = os.path.join(
        os.path.dirname(__file__), "test_metalink_sqlstorage-settings.yml"
    )
    c = get_client(config)
    request = c.mfw_request
    Base.metadata.create_all(bind=request.db_session.bind)

    app = request.app
    col = request.get_collection("tests.object")
    link = request.metalink(col)
    assert link == {
        "type": "morpfw.collection",
        "resource_type": "tests.object",
        "view_name": None,
    }
    assert request.resolve_metalink(link) == "http://localhost:5000/objects"

    obj = col.create({})
    link = request.metalink(obj)
    assert link == {
        "type": "morpfw.model",
        "resource_type": "tests.object",
        "uuid": obj.uuid,
        "view_name": None,
    }
    assert (
        request.resolve_metalink(link) == "http://localhost:5000/objects/%s" % obj.uuid
    )

    with pytest.raises(NotImplementedError):
        request.metalink(object())



# FIXME: this plugin is not testable

import pamela
from .sqlstorage.sqlstorage import UserSQLStorage, APIKeySQLStorage, GroupSQLStorage
from ..user.model import UserModel, UserSchema
from ..group.model import GroupModel, GroupSchema
from ..apikey.model import APIKeyModel, APIKeySchema
from ....crud.errors import UnprocessableError
import socket
from ..app import App


class UserPAMSQLStorage(UserSQLStorage):

    def get(self, identifier):
        try:
            pamela.check_account(identifier)
        except pamela.PAMError:
            return None
        user = super().get(identifier)
        if not user:
            user = self.create({
                'username': identifier,
                'email': '%s@%s' % (identifier, socket.gethostname())
            })
        return user

    def validate(self, userid, password):
        u = self.get_by_userid(userid)
        return pamela.authenticate(username=u.data['username'],
                                   password=password)

    def change_password(self, userid, new_password):
        raise UnprocessableError(
            'Changing password is not possible for PAM auth')



from urllib.parse import urlparse

from .authn.pas.group.model import GroupCollection, GroupSchema
from .authn.pas.path import get_group, get_group_collection
from .authn.pas.path import get_user_collection as get_user_collection


def create_group(request, groupname):
    collection = get_group_collection(request)
    return collection.create({"groupname": groupname})


def delete_group(request, groupname):
    collection = get_group_collection(request)
    group = collection.get_by_groupname(groupname)
    group.delete()


def get_user(request, username):
    collection = get_user_collection(request)
    return collection.get_by_username(username)


def get_user_by_userid(request, userid):
    collection = get_user_collection(request)
    return collection.get_by_userid(userid)





#
from morepath import redirect

from . import crud
from .app import App, BaseApp, SQLApp
from .authn.pas.user.path import get_current_user
from .crud import Collection, Model
from .crud import signals as crudsignals
from .crud.aggregateprovider.base import AggregateProvider
from .crud.blobstorage.fsblobstorage import FSBlobStorage
from .crud.field import Field
from .crud.metalink import MetalinkProvider
from .crud.relationship import BackReference, Reference
from .crud.rulesprovider.base import RulesProvider
from .crud.schema import BaseSchema, Schema
from .crud.searchprovider.base import SearchProvider
from .crud.statemachine.base import StateMachine
from .crud.storage.elasticsearchstorage import ElasticSearchStorage
from .crud.storage.sqlstorage import SQLStorage
from .crud.workflow import WorkflowCollection, WorkflowModel
from .main import create_admin, create_app, run, runprod
from .memoizer import memoize, requestmemoize
from .request import Request, request_factory
from .sql import Base as SQLBase
from .util import get_group, get_user, get_user_by_userid
from .oauth import OAuthRoot



import os

import morepath
import morpfw
import morpfw.sql
import yaml
from sqlalchemy.sql import func, select

from .test_sqlapp import get_pagecollection


def test_sqlsession(pgsql_db):
    morepath.scan(morpfw)
    with open(os.path.join(os.path.dirname(__file__), "test_sqlapp-settings.yml")) as f:
        settings = yaml.load(f.read(), Loader=yaml.Loader)

    with morpfw.request_factory(settings, scan=False) as request:
        morpfw.sql.Base.metadata.create_all(bind=request.db_session.bind)

    with morpfw.request_factory(settings, scan=False) as request:
        col = get_pagecollection(request)
        col.create({"title": "Hello world", "body": "Hello world"})

    with morpfw.request_factory(settings, scan=False) as request:
        col = get_pagecollection(request)
        items = col.search()
        assert len(items) == 1
        items[0].delete()

    with morpfw.request_factory(settings, scan=False) as request:
        col = get_pagecollection(request)
        items = col.search()
        assert len(items) == 0

    # test vacuuming
    with morpfw.request_factory(settings, scan=False) as request:
        col = get_pagecollection(request)
        orm_model = col.storage.orm_model
        session = col.storage.session
        tbl = orm_model.__table__
        sel_stmt = select([func.count(tbl.c.uuid)]).where(orm_model.deleted.isnot(None))
        res = session.execute(sel_stmt)
        assert res.fetchone()[0] == 1

        col.storage.vacuum()

        res = session.execute(sel_stmt)
        assert res.fetchone()[0] == 0

    # ensure clean
    with morpfw.request_factory(settings, scan=False) as request:
        col = get_pagecollection(request)
        orm_model = col.storage.orm_model
        session = col.storage.session
        tbl = orm_model.__table__
        sel_stmt = select([func.count(tbl.c.uuid)]).where(orm_model.deleted.isnot(None))
        res = session.execute(sel_stmt)
        assert res.fetchone()[0] == 0



import code
import cProfile
import os
import sys
import time

import click
import morpfw
import rulez
import transaction

from .cli import cli, load


@cli.command(help="Start MorpFW shell")
@click.option(
    "-e", "--script", required=False, help="Script to run before spawning shell"
)
@click.pass_context
def shell(ctx, script):
    return _start_shell(ctx, script)


@cli.command(help="Profile script")
@click.option("-e", "--script", required=False, help="Script to profile")
@click.pass_context
def profile(ctx, script):
    prof = cProfile.Profile()
    starttime = time.time()
    prof.enable()
    _start_shell(ctx, script, spawn_shell=False)
    prof.disable()
    endtime = time.time()
    print(f"Time taken: {endtime - starttime:.3f} seconds")
    outfile = script + ".pstats"
    if os.path.exists(outfile):
        os.unlink(outfile)
    prof.dump_stats(outfile)
    print(f"Profiler result stored as {outfile}")


@cli.command(help="Execute script")
@click.option("-e", "--script", required=False, help="Script to run")
@click.option(
    "--commit",
    default=False,
    required=False,
    type=bool,
    is_flag=True,
    help="Commit transaction",
)
@click.pass_context
def execute(ctx, script, commit):
    starttime = time.time()
    _start_shell(ctx, script, spawn_shell=False)
    if commit:
        transaction.commit()
    endtime = time.time()
    print(f"Time taken: {endtime - starttime:.3f} seconds")


def _start_shell(ctx, script, spawn_shell=True):
    from morepath.authentication import Identity

    param = load(ctx.obj["settings"])
    settings = param["settings"]
    request = morpfw.request_factory(settings)
    session = request.db_session

    def commit():
        transaction.commit()
        sys.exit()

    localvars = {
        "session": session,
        "request": request,
        "app": request.app,
        "settings": settings,
        "Identity": Identity,
        "commit": commit,
        "morpfw": morpfw,
        "rulez": rulez,
    }
    if script:
        with open(script) as f:
            src = f.read()
            glob = globals().copy()
            filepath = os.path.abspath(script)
            sys.path.insert(0, os.path.dirname(filepath))
            glob["__file__"] = filepath
            bytecode = compile(src, filepath, "exec")
            exec(bytecode, glob, localvars)
    if spawn_shell:
        _shell(localvars)


def _shell(vars):

    # readline is imported here because it screws up with click.prompt
    import readline
    import rlcompleter

    readline.set_completer(rlcompleter.Completer(vars).complete)
    readline.parse_and_bind("tab: complete")
    banner = "\nMorpFW Interactive Console\nAvailable globals : %s\n" % (
        ", ".join(vars.keys())
    )
    sh = code.InteractiveConsole(vars)
    sh.interact(banner=banner)




import warnings
from urllib.parse import urlparse

import dectate
import morepath
import reg
from more.jsonschema import JsonSchemaApp
from more.signals import SignalApp
from sqlalchemy.orm import sessionmaker

from . import component as actions
from . import signals as signals
from .blobstorage.base import NullBlobStorage
from .model import Model

MARKER = object()


class App(JsonSchemaApp, signals.SignalApp):

    dataprovider = dectate.directive(actions.DataProviderAction)
    jsonprovider = dectate.directive(actions.JSONProviderAction)
    formvalidators = dectate.directive(actions.FormValidatorAction)
    identifierfield = dectate.directive(actions.IdentifierFieldAction)
    default_identifier = dectate.directive(actions.DefaultIdentifierAction)
    rulesprovider = dectate.directive(actions.RulesProviderAction)
    uuidfield = dectate.directive(actions.UUIDFieldAction)
    statemachine = dectate.directive(actions.StateMachineAction)
    searchprovider = dectate.directive(actions.SearchProviderAction)
    aggregateprovider = dectate.directive(actions.AggregateProviderAction)
    xattrprovider = dectate.directive(actions.XattrProviderAction)
    storage = dectate.directive(actions.StorageAction)
    storage_factory = dectate.directive(actions.StorageFactoryAction)
    blobstorage = dectate.directive(actions.BlobStorageAction)
    blobstorage_factory = dectate.directive(actions.BlobStorageFactoryAction)
    typeinfo = dectate.directive(actions.TypeInfoFactoryAction)
    metalink = dectate.directive(actions.MetalinkAction)
    authz_rule = dectate.directive(actions.AuthzRuleAction)

    def get_storage(self, model, request):
        blobstorage = self.get_blobstorage(model, request)
        return self._get_storage(model, request, blobstorage)

    def get_blobstorage(self, model, request):
        return self._get_blobstorage(model, request)

    def get_config_blobstorage(self, request: morepath.Request, name: str = None):
        config = self.settings.configuration.__dict__
        if name is None:
            uri = config["morpfw.blobstorage.url"]
        else:
            uri = config["morpfw.blobstorage.url.{}".format(name)]

        scheme = uri.split(":")[0]
        return self._get_blobstorage_factory(scheme)(request=request, uri=uri)

    def get_config_storage(self, request: morepath.Request, name: str = None):
        config = self.settings.configuration.__dict__
        if name is None:
            uri = config["morpfw.storage.url"]
        else:
            uri = config["morpfw.storage.url.{}".format(name)]

        parsed = urlparse(uri)
        return self._get_storage_factory(parsed.scheme)(request, uri)

    @reg.dispatch_method(
        reg.match_class("model"),
        reg.match_instance("request"),
        reg.match_instance("blobstorage"),
    )
    def _get_storage(self, model, request, blobstorage):
        raise NotImplementedError

    @reg.dispatch_method(reg.match_key("name"))
    def _get_storage_factory(self, name):
        raise NotImplementedError

    @reg.dispatch_method(reg.match_class("model"), reg.match_instance("request"))
    def _get_blobstorage(self, model, request):
        return NullBlobStorage()

    @reg.dispatch_method(reg.match_key("name"))
    def _get_blobstorage_factory(self, name):
        raise NotImplementedError

    @reg.dispatch(reg.match_key("name"))
    def get_authz_rule(self, name):
        raise NotImplementedError(name)

    @reg.dispatch_method(
        reg.match_class("schema", lambda self, schema, obj, storage: schema),
        reg.match_instance("obj"),
        reg.match_instance("storage"),
    )
    def get_dataprovider(self, schema, obj, storage):
        raise NotImplementedError(
            "Dataprovider for %s/%s" % (storage.__class__, obj.__class__)
        )

    @reg.dispatch_method(reg.match_instance("obj"))
    def get_jsonprovider(self, obj):
        raise NotImplementedError("JSONProvider for %s" % obj.__class__)

    @reg.dispatch_method(reg.match_class("schema", lambda self, schema: schema))
    def get_formvalidators(self, schema):
        return []

    @reg.dispatch_method(reg.match_class("schema", lambda self, schema: schema))
    def get_identifierfield(self, schema):
        raise NotImplementedError("IdentifierField for %s" % schema)

    @reg.dispatch_method(reg.match_class("schema", lambda self, schema: schema))
    def get_uuidfield(self, schema):
        return "uuid"

    @reg.dispatch_method(
        reg.match_class("schema", lambda self, schema, obj, request: schema)
    )
    def get_default_identifier(self, schema, obj, request):
        return None

    @reg.dispatch_method(reg.match_instance("model", lambda self, context: context))
    def get_rulesprovider(self, context):
        raise NotImplementedError

    @reg.dispatch_method(reg.match_instance("model", lambda self, context: context))
    def get_statemachine(self, context):
        raise NotImplementedError

    @reg.dispatch_method(reg.match_class("model"))
    def get_statemachine_factory(self, model):
        return None

    @reg.dispatch_method(reg.match_instance("model", lambda self, context: context))
    def get_searchprovider(self, context):
        raise NotImplementedError

    @reg.dispatch_method(reg.match_instance("model", lambda self, context: context))
    def get_aggregateprovider(self, context):
        raise NotImplementedError

    @reg.dispatch_method(reg.match_instance("model", lambda self, context: context))
    def get_xattrprovider(self, context):
        raise NotImplementedError

    @reg.dispatch_method(reg.match_key("name"))
    def _get_metalinkprovider_by_name(self, name, request):
        raise NotImplementedError

    @reg.dispatch_method(reg.match_instance("model"))
    def _get_metalinkprovider_by_model(self, model, request):
        raise NotImplementedError

    def get_metalink(self, obj: object, request, **kwargs):
        metalink = self._get_metalinkprovider_by_model(obj, request)
        return metalink.link(obj, **kwargs)

    def resolve_metalink(self, link: dict, request):
        metalink = self._get_metalinkprovider_by_name(link["type"], request)
        return metalink.resolve(link)

    @reg.dispatch_method(reg.match_key("name"))
    def get_typeinfo_factory(self, name):
        raise NotImplementedError

    def get_typeinfo(self, name, request):
        typeinfo = self.config.type_registry.get_typeinfo(name=name, request=request)
        return typeinfo

    def get_typeinfo_by_schema(self, schema, request):
        typeinfo = self.config.type_registry.get_typeinfo_by_schema(
            schema=schema, request=request
        )
        return typeinfo

    def get_compositekey_separator(self):
        raise Exception("BOOO")

    def join_identifier(self, *args):
        raise Exception("BOOO")

    def permits(self, request: morepath.Request, context: Model, permission: str):
        identity = request.identity
        return self._permits(identity, context, permission)

    def get_config(self, key, default=MARKER):
        registry = self.settings.configuration.__dict__
        if default is MARKER:
            return registry.get(key)
        return registry.get(key, default)

    def get_template(self, template):
        def render(content, request):
            return content

        return self.config.template_engine_registry.get_template_render(
            template, render
        )



import os

from ...app import SQLApp
from ...crud.storage.sqlstorage import SQLStorage
from ...sql import Base, construct_orm_model
from ..common import create_admin, get_client, make_request
from .auth_crud import App as BaseApp
from .auth_crud import (
    Object1Model,
    Object1Schema,
    Object2Model,
    Object2Schema,
    run_test,
)


class App(BaseApp, SQLApp):
    pass


App.hook_auth_models()


class Object1Storage(SQLStorage):
    model = Object1Model
    orm_model = construct_orm_model(schema=Object1Schema, metadata=Base.metadata,)


@App.storage(model=Object1Model)
def get_object1_storage(model, request, blobstorage):
    return Object1Storage(request, blobstorage)


class Object2Storage(SQLStorage):
    model = Object2Model
    orm_model = construct_orm_model(schema=Object2Schema, metadata=Base.metadata,)


@App.storage(model=Object2Model)
def get_object2_storage(model, request, blobstorage):
    return Object2Storage(request, blobstorage)


def test_auth_crud(pgsql_db):
    config = os.path.join(
        os.path.dirname(__file__), "test_auth_crud_sqlstorage-settings.yml"
    )
    c = get_client(config)
    request = make_request(c.app)
    Base.metadata.create_all(bind=request.db_session.bind)
    create_admin(c.mfw_request, "admin", "password", "admin@localhost.localdomain")

    run_test(c)



from ..app import App
from ..model import Collection
from .base import AggregateProvider


class StorageAggregateProvider(AggregateProvider):
    def aggregate(self, query=None, group=None, order_by=None, limit=None):
        return self.storage.aggregate(
            query, group=group, order_by=order_by, limit=limit
        )


@App.aggregateprovider(model=Collection)
def get_aggregateprovider(context):
    return StorageAggregateProvider(context)



import abc
from typing import Any, BinaryIO, List, Optional, Sequence, Type, Union

import morepath
import webob

_marker = object()


class ISchema(object):
    """
    Schema represent the fields available in a content type. MorpFW uses
    python dataclass as its schema we provide transformation function to
    transform dataclass into other schema format (eg: JSL, colander, etc)
    """


class IDataProvider(abc.ABC):
    """
    Data provider wraps around backend storage model object
    (eg: SQLAlchemy model object) and provide a bridge for getting and setting
    data from model.
    """

    @abc.abstractmethod
    def __init__(self, schema: Type[ISchema], data: dict, storage: "IStorageBase"):
        super().__init__()

    @abc.abstractmethod
    def __setitem__(self, key, value):
        raise NotImplementedError

    @abc.abstractmethod
    def __getitem__(self, key):
        raise NotImplementedError

    @abc.abstractmethod
    def __delitem__(self, key):
        raise NotImplementedError

    @abc.abstractmethod
    def setdefault(self, key, value):
        raise NotImplementedError

    @abc.abstractmethod
    def get(self, key, default=None):
        raise NotImplementedError

    @abc.abstractmethod
    def set(self, key, value):
        raise NotImplementedError

    @abc.abstractmethod
    def items(self):
        raise NotImplementedError

    @abc.abstractmethod
    def keys(self):
        raise NotImplementedError

    @abc.abstractmethod
    def as_dict(self):
        """Returns dictionary representation of the data, where python objects
        such as ``datetime`` remains as python objects"""
        raise NotImplementedError

    @abc.abstractmethod
    def as_json(self):
        """Returns JSON-safe dictionary representation of the data. Any python
        objects are serialized into JSON-safe data type"""
        raise NotImplementedError

    def _is_json_type(self, value):
        """Check if a value is JSON-safe"""
        if value is None:
            return True
        return (
            isinstance(value, str)
            or isinstance(value, list)
            or isinstance(value, dict)
            or isinstance(value, float)
            or isinstance(value, int)
        )


class IBlob(abc.ABC):
    @abc.abstractmethod
    def __init__(
        self,
        uuid,
        filename: str,
        mimetype: Optional[str] = None,
        size: Optional[int] = None,
        encoding: Optional[str] = None,
    ):
        super().__init__()

    @abc.abstractmethod
    def open(self) -> BinaryIO:
        raise NotImplementedError

    @abc.abstractmethod
    def get_size(self) -> int:
        raise NotImplementedError

    @abc.abstractmethod
    def __call__(self, req: webob.Request) -> webob.Response:
        raise NotImplementedError


class IBlobStorage(abc.ABC):
    @abc.abstractmethod
    def put(
        self,
        field: str,
        fileobj: BinaryIO,
        filename: str,
        mimetype: Optional[str] = None,
        size: Optional[int] = None,
        encoding: Optional[str] = None,
        uuid: Optional[str] = None,
    ) -> IBlob:
        """Receive and store blob data"""
        raise NotImplementedError

    @abc.abstractmethod
    def get(self, uuid: str) -> Optional[IBlob]:
        """Return blob data"""
        raise NotImplementedError

    @abc.abstractmethod
    def delete(self, uuid: str):
        """Delete blob data"""
        raise NotImplementedError


class IStorageBase(abc.ABC):
    """Basic storage backend"""

    @abc.abstractmethod
    def __init__(
        self, request: morepath.Request, blobstorage: Optional[IBlobStorage] = None
    ):
        super().__init__()

    @abc.abstractmethod
    def create(self, data: dict) -> "IModel":
        """Create a model from submitted data"""
        raise NotImplementedError

    @abc.abstractmethod
    def get(self, identifier) -> Optional["IModel"]:
        """return model from identifier"""
        raise NotImplementedError

    @abc.abstractmethod
    def search(
        self,
        query: Optional[dict] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        order_by: Union[None, list, tuple] = None,
    ) -> Sequence["IModel"]:
        """return search result based on specified rulez query"""
        raise NotImplementedError

    @abc.abstractmethod
    def get_by_id(self, id) -> Optional["IModel"]:
        """return model from internal ID"""
        raise NotImplementedError

    @abc.abstractmethod
    def get_by_uuid(self, uuid) -> Optional["IModel"]:
        """return model from uuid"""
        raise NotImplementedError

    @abc.abstractmethod
    def update(self, identifier, data):
        """update model with values from data"""
        raise NotImplementedError

    @abc.abstractmethod
    def delete(self, identifier, model):
        """delete model data"""
        raise NotImplementedError


class IStorage(IStorageBase):
    """Aggregateable storage"""

    @abc.abstractmethod
    def aggregate(
        self,
        query: Optional[dict] = None,
        group: Optional[dict] = None,
        order_by: Union[None, list, tuple] = None,
    ) -> list:
        """return aggregation result based on specified rulez query and group"""
        raise NotImplementedError


class IModel(abc.ABC):
    """
    Model is a representation of a data object. It
    provide a common set of API which is then delegated down
    to the storage provider.

    Model is subscriptable and you can use it like a dictionary
    to access stored data.

    :param request: the request object
    :param storage: storage provider
    :param data: initial data on this model

    """

    #: Set whether object is linkable or not.
    #: If an object is linkable, its json result
    #: will have links attribute
    linkable: bool

    @abc.abstractproperty
    def schema(self) -> Type[ISchema]:
        """The dataclass schema which this model will be using"""
        raise NotImplementedError

    #: When set to True, will enable PATCH view to update model
    update_view_enabled: bool

    #: When set to True, will enable DELETE view to delete model
    delete_view_enabled: bool

    #: Field name on the model data which will be storing blob references
    blobstorage_field: str

    #: List of blob field names allowed on this model
    blob_fields: List[str]

    #: url identifier for this model
    identifier: str

    #: uuid of this model
    uuid: str

    #: Data provider
    data: IDataProvider

    #: List of fields that should be hidden from output
    hidden_fields: list

    @abc.abstractmethod
    def __init__(
        self, request: morepath.Request, collection: "ICollection", data: dict,
    ):
        super().__init__()

    @abc.abstractmethod
    def __getitem__(self, key):
        raise NotImplementedError

    @abc.abstractmethod
    def update(self, newdata: dict, secure: bool):
        """Update model with new data"""
        raise NotImplementedError

    @abc.abstractmethod
    def delete(self):
        """Delete model"""
        raise NotImplementedError

    @abc.abstractmethod
    def save(self):
        """Persist model data into backend storage"""
        raise NotImplementedError

    @abc.abstractmethod
    def json(self) -> dict:
        """Convert model to JSON-safe dictionary"""
        raise NotImplementedError

    @abc.abstractmethod
    def links(self) -> list:
        """Generate links for this model"""
        raise NotImplementedError

    @abc.abstractmethod
    def rulesprovider(self):
        """Return pluggable business rule adapter for this model"""
        raise NotImplementedError

    @abc.abstractmethod
    def statemachine(self):
        """Return PyTransition statemachine adapter for this model"""
        raise NotImplementedError

    @abc.abstractmethod
    def xattrprovider(self):
        """Return extended attributes provider for this model"""
        raise NotImplementedError

    @abc.abstractmethod
    def set_initial_state(self):
        """Initialize default statemachine state for this model"""
        raise NotImplementedError

    @abc.abstractmethod
    def put_blob(
        self,
        field: str,
        fileobj: BinaryIO,
        filename: str,
        mimetype: Optional[str] = None,
        size: Optional[int] = None,
        encoding: Optional[str] = None,
    ) -> IBlob:
        """Receive and store blob object"""
        raise NotImplementedError

    @abc.abstractmethod
    def get_blob(self, field: str) -> IBlob:
        """Return blob"""
        raise NotImplementedError

    @abc.abstractmethod
    def delete_blob(self, field: str):
        """Delete blob"""
        raise NotImplementedError

    # event hooks

    def after_created(self) -> None:
        """Triggered after resource have been created"""

    def before_update(self, newdata: dict) -> None:
        """Triggered before updating resource with new values"""

    def after_updated(self) -> None:
        """Triggered after resource have been created"""

    def before_delete(self) -> bool:
        """
        Triggered before deleting resource
        
        If the return value is False-ish, delete will be prevented
        """
        return True

    def before_blobput(
        self,
        field: str,
        fileobj: BinaryIO,
        filename: str,
        mimetype: Optional[str] = None,
        size: Optional[int] = None,
        encoding: Optional[str] = None,
    ) -> None:
        """Triggered before BLOB is stored"""

    def after_blobput(self, field: str, blob: IBlob) -> None:
        """Triggered after BLOB is stored"""

    def before_blobdelete(self, field: str) -> None:
        """
        Triggered before BLOB is deleted
        
        If the return value is False-ish, delete will be prevented
        """
        return True


class ICollection(abc.ABC):
    """Collection provide an API for querying group of model from its storage"""

    @abc.abstractmethod
    def search(
        self,
        query: Optional[dict] = None,
        offset: int = 0,
        limit: Optional[int] = None,
        order_by: Optional[tuple] = None,
        secure: bool = False,
    ) -> List[IModel]:
        """Search for models

        Filtering is done through ``rulez`` based JSON/dict query, which
        defines boolean statements in JSON/dict structure.

        : param query: Rulez based query
        : param offset: Result offset
        : param limit: Maximum number of result
        : param order_by: Tuple of ``(field, order)`` where ``order`` is
                         ``'asc'`` or ``'desc'``
        : param secure: When set to True, this will filter out any object which
                       current logged in user is not allowed to see

        : todo: ``order_by`` need to allow multiple field ordering
        """
        raise NotImplementedError

    @abc.abstractmethod
    def aggregate(
        self,
        query: Optional[dict] = None,
        group: Optional[dict] = None,
        order_by: Optional[tuple] = None,
    ) -> List[IModel]:
        """Get aggregated results

        : param query: Rulez based query
        : param group: Grouping structure
        : param order_by: Tuple of ``(field, order)`` where ``order`` is
                         ``'asc'`` or ``'desc'``

        : todo: Grouping structure need to be documented

        """
        raise NotImplementedError

    @abc.abstractmethod
    def create(self, data: dict) -> IModel:
        """Create a model from data"""
        raise NotImplementedError

    @abc.abstractmethod
    def get(self, identifier) -> IModel:
        """Get model by url identifier key"""
        raise NotImplementedError

    @abc.abstractmethod
    def get_by_uuid(self, uuid: str) -> IModel:
        """Get model by uuid"""
        raise NotImplementedError

    @abc.abstractmethod
    def json(self) -> dict:
        """JSON-safe dictionary representing this collection"""
        raise NotImplementedError

    @abc.abstractmethod
    def links(self) -> list:
        """Links related to this collection"""
        raise NotImplementedError

    def before_create(self, data: dict) -> None:
        """Triggered before the creation of resource"""


class IStateMachine(abc.ABC):
    @abc.abstractproperty
    def states(self) -> List:
        """List of ``pytransitions`` states"""
        raise NotImplementedError

    @abc.abstractproperty
    def readonly_states(self) -> List:
        """List of readonly states"""
        raise NotImplementedError

    @abc.abstractproperty
    def transitions(self) -> List:
        """List of ``pytransitions`` transitions"""
        raise NotImplementedError

    @abc.abstractmethod
    def _get_state(self) -> Optional[str]:
        """Get current state"""
        raise NotImplementedError

    @abc.abstractmethod
    def _set_state(self, val):
        """Set current state"""
        raise NotImplementedError

    state = property(_get_state, _set_state, doc="Current resource state")

    @abc.abstractmethod
    def get_triggers(self) -> List:
        """Returns list of available triggers"""
        raise NotImplementedError


class IXattrProvider(abc.ABC):
    @abc.abstractproperty
    def schema(self) -> Type[Any]:
        """Schema to use for data validation"""
        raise NotImplementedError

    # Set to True to allow arbitrary properties besides
    # the properties defined on the schema
    additional_properties: bool = False

    @abc.abstractmethod
    def jsonschema(self) -> dict:
        """Returns JSON Schema for Xattr"""
        raise NotImplementedError

    @abc.abstractmethod
    def as_dict(self):
        """Returns dictionary representation of the data, where python objects
        such as ``datetime`` remains as python objects"""
        raise NotImplementedError

    @abc.abstractmethod
    def as_json(self):
        """Returns JSON-safe dictionary representation of the data. Any python
        objects are serialized into JSON-safe data type"""
        raise NotImplementedError

    @abc.abstractmethod
    def process_update(self, newdata: dict):
        """Validate received data and then update the extended attributes"""
        raise NotImplementedError

    @abc.abstractmethod
    def update(self, newdata: dict):
        """Update extended attributes"""
        raise NotImplementedError

    @abc.abstractmethod
    def __setitem__(self, key, value):
        raise NotImplementedError

    @abc.abstractmethod
    def __getitem__(self, key):
        raise NotImplementedError

    @abc.abstractmethod
    def __delitem__(self, key):
        raise NotImplementedError

    @abc.abstractmethod
    def get(self, key, default=_marker):
        raise NotImplementedError


class ISearchProvider(abc.ABC):
    @abc.abstractmethod
    def parse_query(self, qs: str) -> dict:
        """Parse query string from search query and convert it into rulez query
        dictionary. The dictionary would be passed to search method afterwards,
        which you can then parse into your backend search query.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def search(
        self,
        query: Optional[dict] = None,
        offset: int = 0,
        limit: Optional[int] = None,
        order_by: Union[None, list, tuple] = None,
    ) -> List[IModel]:
        """Execute search and return list of model objects"""
        raise NotImplementedError


class IAggregateProvider(abc.ABC):
    @abc.abstractmethod
    def parse_query(self, qs: str) -> dict:
        """Parse query string from search query and convert it into rulez query
        dictionary. The dictionary would be passed to aggregate method afterwards,
        which you can then parse into your backend search query.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def parse_group(self, qs: str) -> dict:
        """Parse query string from group query and convert it into dictionary
        representation. The dictionary would be passed to aggregate method afterwards,
        which you can then parse into your backend aggregate query.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def aggregate(
        self,
        query: Optional[dict] = None,
        group: Optional[dict] = None,
        order_by: Union[None, list, tuple] = None,
    ) -> list:
        """return aggregation result based on specified rulez query and group"""
        raise NotImplementedError



import copy
import datetime
import uuid
from dataclasses import _MISSING_TYPE

import pytz

import morepath
import sqlalchemy as sa
import sqlalchemy_jsonfield as sajson
from dateutil.parser import parse as _parse_date
from inverter import dc2colanderjson
from inverter.common import dataclass_get_type
from morpfw.authn.pas.policy import Identity

from ...interfaces import IDataProvider, ISchema
from ..app import App
from ..storage.sqlstorage import GUID, Base, MappedTable, SQLStorage
from ..types import datestr
from .dictprovider import DictProvider

_MARKER: list = []


def parse_date(datestr):
    d = _parse_date(datestr)
    if d.tzinfo is None:
        d = d.replace(tzinfo=pytz.UTC)
    d = d.astimezone(pytz.UTC)
    return d.replace(tzinfo=None)


class SQLAlchemyModelProvider(IDataProvider):
    def __init__(self, schema, data, storage):
        self.request = storage.request
        self.schema = schema
        self.data = data
        self.orm_model = data.__class__
        self.columns = self.orm_model.__table__.c
        self.storage = storage
        self.changed = False

    def __getitem__(self, key):
        if isinstance(self.columns[key].type, sa.DateTime):
            try:
                data = getattr(self.data, key)
            except AttributeError:
                raise KeyError(key)
            if data:
                data = data.astimezone(self.request.timezone())
                return data
            return None
        if isinstance(self.columns[key].type, GUID):
            try:
                data = getattr(self.data, key)
            except AttributeError:
                raise KeyError(key)
            if data:
                return data.hex
            return None
        if isinstance(self.columns[key].type, sajson.JSONField):
            try:
                data = getattr(self.data, key)
            except AttributeError:
                raise KeyError(key)
            if data is not None:
                return copy.deepcopy(data)
            return None
        try:
            return getattr(self.data, key)
        except AttributeError:
            raise KeyError(key)

    def __setitem__(self, key, value):
        if key not in self.columns:
            return
        if value and isinstance(self.columns[key].type, sa.DateTime):
            if not isinstance(value, datetime.datetime):
                value = parse_date(value)
            value = value.astimezone(pytz.UTC)
        elif value and isinstance(self.columns[key].type, GUID):
            value = uuid.UUID(value)

        setattr(self.data, key, value)

    def __delitem__(self, key):
        setattr(self.data, key, None)

    def setdefault(self, key, value):
        if self.get(key):
            return
        self.set(key, value)

    def get(self, key, default=_MARKER):
        if default is _MARKER:
            attr = self.schema.__dataclass_fields__.get(key, None)
            if attr:
                # t = dataclass_get_type(attr)
                if not isinstance(attr.default, _MISSING_TYPE):
                    default = attr.default
                elif not isinstance(attr.default_factory, _MISSING_TYPE):
                    default = attr.default_factory()
                else:
                    default = None

        if default is not _MARKER:
            try:
                return self[key]
            except KeyError:
                return default
        return self[key]

    def set(self, key, value):
        self[key] = value

    def items(self):
        res = []
        fields = self.schema.__dataclass_fields__.keys()
        for f in fields:
            res.append((f, self.data[f]))
        return res

    def keys(self):
        return self.schema.__dataclass_fields__.keys()

    def as_dict(self):
        fields = self.schema.__dataclass_fields__.items()
        result = {}
        for n, f in fields:
            v = self.get(n)
            t = dataclass_get_type(f)
            if not v and t["metadata"]["exclude_if_empty"]:
                continue
            result[n] = v
        return result

    def as_json(self):
        fields = self.schema.__dataclass_fields__.items()
        result = {}
        for n, f in fields:
            t = dataclass_get_type(f)
            v = self.get(n)
            if v is None and t["metadata"]["exclude_if_empty"]:
                continue
            result[n] = v
        cschema = dc2colanderjson.convert(self.schema, request=self.storage.request)
        result = cschema().serialize(result)
        return result


@App.dataprovider(schema=ISchema, obj=MappedTable, storage=SQLStorage)
def get_provider(schema, obj, storage):
    return SQLAlchemyModelProvider(schema, obj, storage)


@App.dataprovider(schema=ISchema, obj=Base, storage=SQLStorage)
def get_provider(schema, obj, storage):
    return SQLAlchemyModelProvider(schema, obj, storage)


@App.dataprovider(schema=ISchema, obj=dict, storage=SQLStorage)
def get_dict_provider(schema, obj, storage):
    return DictProvider(schema, obj, storage)


@App.jsonprovider(obj=SQLAlchemyModelProvider)
def get_jsonprovider(obj):
    return obj.as_json()



import re
import secrets
from datetime import datetime
from uuid import uuid4

import morepath
import pytz
import sqlalchemy as sa
import sqlalchemy_jsonfield as sajson
from morpfw.crud import Collection, Model, StateMachine
from morpfw.crud import errors as cruderrors
from morpfw.crud import signals as crudsignal
from morpfw.crud.schema import Schema
from morpfw.crud.validator import regex_validator

from .. import exc
from ..app import App
from ..exc import UserExistsError
from ..group.model import GroupCollection, GroupModel, GroupSchema
from ..group.path import get_group_collection
from ..model import EMAIL_PATTERN, NAME_PATTERN
from ..utils import has_role
from .schema import LoginSchema, RegistrationSchema, UserSchema


class UserCollection(Collection):
    schema = UserSchema
    exist_exc = UserExistsError
    create_view_enabled = False

    def authenticate(self, username, password):
        if re.match(EMAIL_PATTERN, username):
            user = self.storage.get_by_email(self, username)
        else:
            user = self.storage.get_by_username(self, username)

        if user is None:
            return None
        if not user.validate(password):
            return None
        return user

    def get_by_userid(self, userid):
        return self.storage.get_by_userid(self, userid)

    def get_by_email(self, email):
        return self.storage.get_by_email(self, email)

    def get_by_username(self, username):
        return self.storage.get_by_username(self, username)

    def _create(self, data):
        return super(UserCollection, self)._create(data)


class UserModel(Model):

    schema = UserSchema

    blob_fields = ["profile-photo"]

    def title(self):
        return self["username"]

    @property
    def userid(self):
        return self.storage.get_userid(self)

    @property
    def identity(self):
        return morepath.Identity(self.userid)

    def change_password(self, password: str, new_password: str, secure: bool = True):
        rules = self.rulesprovider()
        result = rules.change_password(password, new_password, secure=secure)
        self.update({"nonce": secrets.token_hex(8)}, deserialize=False)
        return result

    def validate(self, password: str, check_state: bool = True):
        rules = self.rulesprovider()
        return rules.validate(password, check_state)

    def groups(self):
        return self.storage.get_user_groups(self.collection, self.userid)

    def group_roles(self):
        group_roles = {}
        for g in self.groups():
            group_roles[g.data["groupname"]] = g.get_member_roles(self.userid)
        return group_roles

    def timezone(self):
        tz = self["timezone"] or "UTC"
        return pytz.timezone(tz).localize(datetime.now()).tzinfo

    def _links(self):
        links = super()._links()
        links.append({"rel": "userid", "value": self.userid})
        for g in self.groups():
            links.append(
                {
                    "rel": "group",
                    "uuid": g.identifier,
                    "name": g.data["groupname"],
                    "href": self.request.link(g),
                }
            )
        return links


class UserStateMachine(StateMachine):

    states = ["active", "inactive", "deleted"]
    transitions = [
        {"trigger": "activate", "source": "inactive", "dest": "active"},
        {"trigger": "deactivate", "source": "active", "dest": "inactive"},
    ]


@App.statemachine(model=UserModel)
def userstatemachine(context):
    return UserStateMachine(context)


@App.subscribe(signal=crudsignal.OBJECT_CREATED, model=UserModel)
def add_user_to_default_group(app, request, obj, signal):
    request = obj.request
    gcol = get_group_collection(request)
    g = gcol.get_by_groupname("__default__")
    if g is None:
        g = gcol.create({"groupname": "__default__"})
    g.add_members([obj.userid])


class CurrentUserModel(UserModel):
    pass



import socket
import threading
from datetime import datetime

import click

from .cli import cli, load


def start_worker(ctx):
    print(threading.get_ident())
    param = load(ctx.obj["settings"])
    hostname = socket.gethostname()
    ws = param["settings"]["configuration"]["morpfw.celery"]
    now = datetime.utcnow().strftime(r"%Y%m%d%H%M")
    # scan
    param["factory"](param["settings"], instantiate=False)
    w = param["class"].celery.Worker(hostname="worker%s.%s" % (now, hostname), **ws)
    w.start()


@cli.command(help="alias to 'worker'")
@click.pass_context
def solo_worker(ctx):
    return start_worker(ctx)


@cli.command(help="start celery worker")
@click.pass_context
def worker(ctx):
    return start_worker(ctx)




import morepath

from ..authn.pas import permission as authperm
from ..authn.pas.apikey.model import APIKeyCollection, APIKeyModel
from ..authn.pas.group.model import GroupCollection, GroupModel
from ..authn.pas.group.path import get_group_collection
from ..authn.pas.group.schema import GroupSchema
from ..authn.pas.user.model import CurrentUserModel, UserCollection, UserModel
from ..authn.pas.utils import has_role
from ..crud import permission as crudperm
from ..crud.model import Collection, Model


class DefaultAuthzPolicy(morepath.App):
    pass


def _has_admin_role(context):
    return has_role(context.request, "administrator")





from .model import APIKeyCollection, APIKeyModel
from .schema import APIKeySchema
from .path import get_apikey, get_apikey_collection
from ..app import App


@App.typeinfo(name='morpfw.pas.apikey', schema=APIKeySchema)
def get_typeinfo(request):
    return {
        'title': 'API Key',
        'description': 'API Key type',
        'schema': APIKeySchema,
        'collection': APIKeyCollection,
        'collection_factory': get_apikey_collection,
        'model': APIKeyModel,
        'model_factory': get_apikey,
        'internal': True,
    }



import os
import shutil
import socket
import time
from contextlib import closing

import mirakuru
import mirakuru.exceptions
import pika
import pytest
from elasticsearch import Elasticsearch
from pytest_postgresql import factories
from pytest_rabbitmq.factories import rabbitmq_proc

pgsql_proc = factories.postgresql_proc(
    executable="/usr/bin/pg_ctl", host="localhost", port=45678, user="postgres"
)
pgsql_db = factories.postgresql("pgsql_proc", db_name="morp_tests")

rabbitmq_pika_proc = rabbitmq_proc(
    server="/usr/lib/rabbitmq/bin/rabbitmq-server", port=34567
)


@pytest.fixture
def pika_connection_channel(rabbitmq_pika_proc):
    connection = pika.BlockingConnection(
        pika.ConnectionParameters(
            host="localhost",
            port=34567,
            credentials=pika.PlainCredentials("guest", "guest"),
        )
    )
    channel = connection.channel()
    return [connection, channel]


@pytest.fixture(scope="session")
def es_proc(request):
    port = 9085
    home_dir = "/tmp/elasticsearch_%s" % port
    os.environ["ES_HOME"] = home_dir
    command = [
        os.environ["ELASTICSEARCH_EXECUTABLE"],
        "-p",
        "/tmp/elasticsearch.%s.pid" % port,
        "-E",
        "http.port=%s" % port,
        "-E",
        "path.logs=/tmp/elasticsearch_%s_logs" % port,
        "-E",
        "cluster.name=elasticsearch_cluster_%s" % port,
        "-E",
        "network.publish_host=127.0.0.1",
        "-E",
        "index.store.type=mmapfs",
    ]
    es_proc = mirakuru.HTTPExecutor(
        command, shell=True, url="http://127.0.0.1:%s" % port
    )
    es_proc.start()

    def finalize_elasticsearch():
        try:
            es_proc.stop()
        except mirakuru.exceptions.ProcessFinishedWithError as e:
            if e.exit_code != 143:
                raise e
        if os.path.exists(home_dir):
            shutil.rmtree(home_dir)

    request.addfinalizer(finalize_elasticsearch)
    return es_proc


@pytest.fixture(scope="session")
def es_client(request):
    process = request.getfixturevalue("es_proc")
    if not process.running():
        process.start()

    hosts = "%s:%s" % (process.host, process.port)

    client = Elasticsearch(hosts=hosts)

    def drop_indexes():
        client.indices.delete(index="*")

    request.addfinalizer(drop_indexes)

    return client


@pytest.fixture(scope="function")
def oauth_server(request):
    port = 5000
    command = [
        os.path.join(os.environ["BUILDOUT_BINDIR"], "morpfw"),
        "-s",
        os.path.join(os.path.dirname(__file__), "auth_test", "test_oauth-settings.yml"),
        "start",
    ]
    oauth_proc = mirakuru.HTTPExecutor(
        command, shell=True, url="http://127.0.0.1:%s" % port
    )
    oauth_proc.start()

    def finalize():
        oauth_proc.stop()

    request.addfinalizer(finalize)
    return oauth_proc



import importlib

from .authn.pas import permission as authperm
from .authn.pas.apikey.model import APIKeyCollection, APIKeyModel
from .authn.pas.user.model import CurrentUserModel, UserCollection, UserModel
from .authz.pas import DefaultAuthzPolicy
from .crud import permission as crudperms
from .crud.model import Collection, Model
from .permission import All

Policy = DefaultAuthzPolicy


def import_name(name):
    modname, objname = name.split(":")
    mod = importlib.import_module(modname)
    return getattr(mod, objname)


def eval_config_groupperms(request, model, permission, identity):
    usercol = request.get_collection("morpfw.pas.user")
    user = usercol.get_by_userid(identity.userid)
    if user["is_administrator"]:
        return True

    if isinstance(model, (Collection, Model)):
        config = request.app.get_config("morpfw.authz.type_permissions", {})
        typeinfo = request.app.get_typeinfo_by_schema(model.schema, request)
        type_conf = config.get(typeinfo["name"], {})
        for g in user.groups():
            rule_name = type_conf.get(g["groupname"], None)
            if rule_name:
                rule_func = request.app.get_authz_rule(rule_name)
                if rule_func(g["groupname"], identity, model, permission):
                    return True

    config = request.app.get_config("morpfw.authz.model_permissions", {})

    model_conf = None
    for model_name in config.keys():
        model_klass = import_name(model_name)
        if isinstance(model, model_klass):
            model_conf = config[model_name]
            break

    if not model_conf:
        return

    for g in user.groups():
        rule_name = model_conf.get(g["groupname"], None)
        if rule_name:
            rule_func = request.app.get_authz_rule(rule_name)
            if rule_func(g["groupname"], identity, model, permission):
                return True

    return None


def currentuser_permission(identity, model, permission):
    request = model.request
    usercol = request.get_collection("morpfw.pas.user")
    user = usercol.get_by_userid(identity.userid)
    if user["is_administrator"]:
        return True
    userid = identity.userid
    if isinstance(model, UserModel):
        if model.userid == userid:
            return True
    elif isinstance(model, APIKeyModel):
        if model["userid"] == userid:
            return True

    return False


@Policy.permission_rule(model=UserCollection, permission=authperm.Register)
def allow_api_registration(identity, model, permission):
    return model.request.app.get_config("morpfw.new_registration.enabled", True)


@Policy.permission_rule(model=UserModel, permission=crudperms.All)
def allow_user_crud(identity, model, permission):
    return currentuser_permission(identity, model, permission)


@Policy.permission_rule(model=UserModel, permission=authperm.ChangePassword)
def allow_change_all_user_password(identity, model, permission):
    request = model.request
    usercol = request.get_collection("morpfw.pas.user")
    user = usercol.get_by_userid(identity.userid)
    if user["is_administrator"]:
        return True
    return False


@Policy.permission_rule(model=CurrentUserModel, permission=authperm.ChangePassword)
def allow_change_self_password(identity, model, permission):
    return currentuser_permission(identity, model, permission)


@Policy.permission_rule(model=APIKeyCollection, permission=crudperms.All)
def allow_apikeycollection_management(identity, model, permission):
    request = model.request
    usercol = request.get_collection("morpfw.pas.user")
    user = usercol.get_by_userid(identity.userid)
    if user:
        return True
    return False


@Policy.permission_rule(model=APIKeyModel, permission=crudperms.All)
def allow_apikey_management(identity, model, permission):
    return currentuser_permission(identity, model, permission)


@Policy.permission_rule(model=Collection, permission=All)
def collection_permission(identity, model, permission):
    return eval_config_groupperms(model.request, model, permission, identity)


@Policy.permission_rule(model=Model, permission=All)
def model_permission(identity, model, permission):
    return eval_config_groupperms(model.request, model, permission, identity)




import abc
import morepath
from typing import Optional, Type


class AuthnPolicy(abc.ABC):

    app_cls: Optional[Type[morepath.App]] = None

    def __init__(self, policy_settings):
        self.policy_settings = policy_settings
        super().__init__()

    @abc.abstractmethod
    def get_identity_policy(self, settings):
        raise NotImplementedError

    @abc.abstractmethod
    def verify_identity(self, app, identity):
        raise NotImplementedError



import json
import logging
import os
import sys
import traceback
from urllib.parse import urlencode

from jsonpath_ng import parse as jsonpath_parse
from morepath.request import Request
from transitions import MachineError
from webob.exc import HTTPForbidden, HTTPInternalServerError, HTTPNotFound

from . import permission
from .app import App
from .errors import (
    AlreadyExistsError,
    FieldValidationError,
    StateUpdateProhibitedError,
    UnprocessableError,
    ValidationError,
)
from .model import Collection, Model
from .validator import get_data, validate_schema

logger = logging.getLogger("morp")


@get_data.register(model=Collection, request=Request)
def get_collection_data(model, request):
    return request.json


@App.json(model=Collection, permission=permission.View)
def schema(context, request):
    return context.json()


@App.json(model=Collection, name="aggregate", permission=permission.Aggregate)
def aggregate(context, request):
    if not context.aggregate_view_enabled:
        raise HTTPNotFound()

    qs = request.GET.get("q", "").strip()

    query = None

    aggregateprovider = context.aggregateprovider()

    if qs:
        query = aggregateprovider.parse_query(qs)

    order_by = request.GET.get("order_by", None)
    limit = request.GET.get("limit", None)

    gs = request.GET.get("group", "").strip()
    group = None
    if gs:
        group = aggregateprovider.parse_group(gs)

    if order_by:
        order_by = order_by.split(":")
        if len(order_by) == 1:
            order_by = order_by + ["asc"]
    objs = context.aggregate(query, group=group, order_by=order_by, limit=limit)
    return objs


@App.json(model=Collection, name="search", permission=permission.Search)
def search(context, request):
    if not context.search_view_enabled:
        raise HTTPNotFound()

    qs = request.GET.get("q", "").strip()

    query = None
    if qs:
        searchprovider = context.searchprovider()
        query = searchprovider.parse_query(qs)

    default_limit = request.app.get_config("morpfw.crud.search.default_limit", 20)
    max_limit = request.app.get_config("morpfw.crud.search.max_limit", 100)
    limit = int(request.GET.get("limit", 0)) or default_limit
    if limit > max_limit:
        limit = max_limit
    offset = int(request.GET.get("offset", 0))
    order_by = request.GET.get("order_by", None)
    select = request.GET.get("select", None)
    if order_by:
        order_by = order_by.split(":")
        if len(order_by) == 1:
            order_by = order_by + ["asc"]
    # HACK: +1 to ensure next page links is triggered
    searchlimit = limit
    if limit:
        searchlimit = limit + 1
    objs = context.search(
        query, offset=offset, limit=searchlimit, order_by=order_by, secure=True
    )
    # and limit back to actual limit
    objs = [obj.json() for obj in objs]
    has_next = False
    if limit:
        if len(objs) > limit:
            has_next = True
        objs = objs[:limit]
    if select:
        expr = jsonpath_parse(select)
        results = []
        for obj in objs:
            results.append([match.value for match in expr.find(obj["data"])])
    else:
        results = objs
    params = {}
    if select:
        params["select"] = select
    if limit:
        params["limit"] = limit
    if order_by:
        params["order_by"] = request.GET.get("order_by", "")
    params["offset"] = offset + (limit or 0)
    qs = urlencode(params)
    res = {
        "results": results,
        "q": query,
        "limit": limit,
        "order_by": order_by,
        "offset": offset,
        "result_count": len(results),
    }
    if has_next:
        res.setdefault("links", [])
        res["links"].append(
            {"rel": "next", "href": request.link(context, "+search?%s" % qs)}
        )
    if offset > 0:
        prev_offset = offset - (limit or 0)
        if prev_offset < 0:
            prev_offset = 0
        params["offset"] = prev_offset
        qs = urlencode(params)
        res.setdefault("links", [])
        res["links"].append(
            {"rel": "previous", "href": request.link(context, "+search?%s" % qs)}
        )
    return res


@App.json(model=Collection, request_method="POST", permission=permission.Create)
def create(context, request):
    if not context.create_view_enabled:
        raise HTTPNotFound()
    obj = context.create(request.json, secure=True)
    return obj.json()


@get_data.register(model=Model, request=Request)
def get_obj_data(model, request):
    data = model.json()["data"]
    data.update(request.json)
    return data


@App.json(model=Model, permission=permission.View)
def read(context, request):
    select = request.GET.get("select", None)
    obj = context.json()
    if select:
        expr = jsonpath_parse(select)
        obj = [match.value for match in expr.find(obj["data"])]
    return obj


@App.json(model=Model, request_method="PATCH", permission=permission.Edit)
def update(context, request):
    if not context.update_view_enabled:
        raise HTTPNotFound()

    context.update(request.json, secure=True)
    return {"status": "success"}


@App.json(
    model=Model,
    name="statemachine",
    request_method="POST",
    permission=permission.StateUpdate,
)
def statemachine(context, request):
    if not context.statemachine_view_enabled:
        raise HTTPNotFound()

    sm = context.statemachine()
    transition = request.json["transition"]
    transition_callable = getattr(sm, transition, None)
    if transition_callable and (transition not in sm.get_triggers()):

        @request.after
        def adjust_status(response):
            response.status = 422

        return {
            "status": "error",
            "message": "Transition %s is not allowed" % transition,
        }

    if not transition_callable:

        @request.after
        def adjust_status(response):
            response.status = 422

        return {"status": "error", "message": "Unknown transition %s" % transition}
    transition_callable()
    context.save()
    return {"status": "success"}


@App.json(model=Model, request_method="DELETE", permission=permission.Delete)
def delete(context, request):
    if not context.delete_view_enabled:
        raise HTTPNotFound()

    context.delete()
    return {"status": "success"}


@App.view(model=Model, name="blobs", permission=permission.View)
def get_blob(context, request):
    field = request.GET.get("field", None)
    if field is None or context.storage.blobstorage is None:
        raise HTTPNotFound()

    if field not in context.blob_fields:
        raise HTTPNotFound()

    blob = context.get_blob(field)

    if blob is None:
        raise HTTPNotFound()

    return request.get_response(blob)


@App.json(model=Model, name="blobs", request_method="POST", permission=permission.Edit)
def put_blob(context, request):
    field = request.GET.get("field", None)
    if field is None or context.storage.blobstorage is None:
        raise HTTPNotFound()

    if field not in context.blob_fields:
        raise HTTPNotFound()

    upload = request.POST.get("upload")

    if upload is None:
        raise UnprocessableError()

    context.put_blob(
        field,
        filename=os.path.basename(upload.filename),
        mimetype=upload.type,
        fileobj=upload.file,
    )
    return {"status": "success"}


@App.json(
    model=Model, name="blobs", request_method="DELETE", permission=permission.Edit
)
def delete_blob(context, request):
    field = request.GET.get("field", None)
    if field is None or context.storage.blobstorage is None:
        raise HTTPNotFound()

    if field not in context.blob_fields:
        raise HTTPNotFound()

    context.delete_blob(field)

    return {"status": "success"}


@App.json(model=Model, name="xattr-schema", permission=permission.View)
def get_xattr_schema(context, request):
    if not context.xattr_view_enabled:
        raise HTTPNotFound()

    provider = context.xattrprovider()
    return provider.jsonschema()


@App.json(model=Model, name="xattr", permission=permission.View)
def get_xattr(context, request):
    if not context.xattr_view_enabled:
        raise HTTPNotFound()

    provider = context.xattrprovider()
    return provider.as_json()


@App.json(model=Model, name="xattr", permission=permission.View, request_method="PATCH")
def set_xattr(context, request):
    if not context.xattr_view_enabled:
        raise HTTPNotFound()

    provider = context.xattrprovider()
    provider.process_update(request.json)
    return {"status": "success"}


@App.json(model=AlreadyExistsError)
def alreadyexists_error(context, request):
    @request.after
    def adjust_status(response):
        response.status = 422

    #   FIXME: should log this when a config for DEBUG_SECURITY is enabled
    #    logger.error(traceback.format_exc())
    #    print(traceback.format_exc())

    return {
        "status": "error",
        "message": "Object Already Exists : %s" % context.message,
    }


@App.json(model=HTTPForbidden)
def forbidden_error(context, request):
    @request.after
    def adjust_status(response):
        response.status = 403

    #   FIXME: should log this when a config for DEBUG_SECURITY is enabled
    #    logger.error(traceback.format_exc())
    return {"status": "error", "message": "Access Denied : %s" % request.path}


@App.json(model=HTTPNotFound)
def httpnotfound_error(context, request):
    @request.after
    def adjust_status(response):
        response.status = 404

    return {"status": "error", "message": "Object Not Found : %s" % request.path}


@App.json(model=ValidationError)
def validation_error(context, request):
    @request.after
    def adjust_status(response):
        response.status = 422

    field_errors = []
    form_errors = []
    for e in context.field_errors:
        field_errors.append({"field": e.path, "message": e.message})

    for e in context.form_errors:
        form_errors.append(e.message)

    #   FIXME: should log this when a config for DEBUG_SECURITY is enabled
    # logger.error(traceback.format_exc())

    return {"status": "error", "field_errors": field_errors, "form_errors": form_errors}


@App.json(model=FieldValidationError)
def field_validation_error(context, request):
    @request.after
    def adjust_status(response):
        response.status = 422

    form_errors = [context.args[0]]

    return {"status": "error", "field_errors": [], "form_errors": form_errors}


@App.json(model=MachineError)
def statemachine_error(context, request):
    @request.after
    def adjust_status(response):
        response.status = 422

    return {"status": "error", "message": context.value}


@App.json(model=StateUpdateProhibitedError)
def stateupdateprohibited_error(context, request):
    @request.after
    def adjust_status(response):
        response.status = 422

    return {
        "status": "error",
        "message": "Please use 'statemachine' view to update state",
    }


@App.json(model=Exception)
def internalserver_error(context, request):
    @request.after
    def adjust_status(response):
        response.status = 500

    tb = traceback.format_exc()
    logger.error(tb)

    return {
        "status": "error",
        "message": "Internal server error",
        "traceback": tb.split("\n"),
    }


@App.json(model=UnprocessableError)
def unprocessable_error(context, request):
    @request.after
    def adjust_status(response):
        response.status = 422

    return {"status": "error", "message": "%s" % (context.message)}



import click
import morpfw

from .cli import cli, load


@cli.command(help="register administrator user (only for app using PAS)")
@click.option("-u", "--username", required=True, help="Username", prompt=True)
@click.option("-e", "--email", required=True, help="Email address", prompt=True)
@click.option(
    "-p",
    "--password",
    required=True,
    help="Password",
    prompt=True,
    hide_input=True,
    confirmation_prompt=True,
)
@click.pass_context
def register_admin(ctx, username, email, password):
    param = load(ctx.obj["settings"])
    settings = param["settings"]
    with morpfw.request_factory(
        settings, extra_environ={"morpfw.nomemoize": True}
    ) as request:
        user = morpfw.create_admin(
            request, username=username, password=password, email=email
        )
        if user is None:
            print("Application is not using Pluggable Auth Service")




import os
import sys

from sqlalchemy import MetaData, create_engine
from sqlalchemy import inspect as Inspector
from sqlalchemy.schema import DropSchema

from alembic import command
from alembic.config import Config


def _load_engines(config):
    engines = {}
    for k, v in config.items():
        k = k.strip()
        prefix = "morpfw.storage.sqlstorage.dburl"
        if k.startswith(prefix):
            name = k.replace(prefix, "").strip()
            if name == "":
                name = "default"
            elif name.startswith("."):
                name = name[1:]

            engines[name] = {"dburl": v}
    return engines


def drop_all(request):

    config = request.app.settings.configuration.__dict__
    engines = _load_engines(config)

    for n, conf in engines.items():
        param = dict([(k, v) for k, v in conf.items() if k != "dburl"])
        engine = create_engine(conf["dburl"], **param)
        inspect = Inspector(engine)
        print("Clearing engine: %s" % n)
        for schema in inspect.get_schema_names():
            if schema in ["information_schema"]:
                continue
            if schema.startswith("pg_"):
                continue
            meta = MetaData(engine, schema=schema)
            print(".. Clearing tables from schema {}".format(schema))
            meta.reflect()
            meta.drop_all()
            if schema in ["public", "dbo"]:
                continue

            print("Dropping schema {}".format(schema))
            engine.execute(DropSchema(schema))
        print("Done")



import os
import typing

from webob import exc
from webob.dec import wsgify
from webob.response import Response
from webob.static import FileIter

from ...interfaces import IBlob, IBlobStorage

BLOCK_SIZE = 1 << 16


class Blob(IBlob):
    def __init__(
        self, uuid, filename, mimetype=None, size=None, encoding=None, sha256sum=None
    ):
        self.uuid = uuid
        if size and size < 0:
            size = None
        self.size = size
        self.filename = filename
        self.mimetype = mimetype
        self.encoding = encoding
        self.sha256sum = sha256sum
        super().__init__(uuid, filename, mimetype, size, encoding)

    def get_size(self) -> int:
        return self.size

    @wsgify
    def __call__(self, req):
        if req.method not in ("GET", "HEAD"):
            return exc.HTTPMethodNotAllowed("You cannot %s a file" % req.method)

        disposition = req.GET.get("disposition", "attachment")
        try:
            file = self.open()
        except (IOError, OSError) as e:
            msg = "You are not permitted to view this file (%s)" % e
            return exc.HTTPForbidden(msg)

        if "wsgi.file_wrapper" in req.environ:
            app_iter = req.environ["wsgi.file_wrapper"](file, BLOCK_SIZE)
        else:
            app_iter = FileIter(file)

        return Response(
            app_iter=app_iter,
            content_length=self.get_size(),
            content_type=self.mimetype,
            content_encoding=self.encoding,
            content_disposition='%s; filename="%s"' % (disposition, self.filename),
            # @@ etag
        ).conditional_response_app


class BlobStorage(IBlobStorage):
    pass


class NullBlobStorage(BlobStorage):
    def put(
        self,
        fileobj: typing.BinaryIO,
        filename: str,
        mimetype: typing.Optional[str] = None,
        size: typing.Optional[int] = None,
        encoding: typing.Optional[str] = None,
        uuid: typing.Optional[str] = None,
    ) -> Blob:
        raise NotImplementedError

    def get(self, uuid: str) -> typing.Optional[Blob]:
        raise NotImplementedError

    def delete(self, uuid: str):
        raise NotImplementedError



from .app import App
import morepath
import os
import sqlalchemy
from morpfw.crud.storage.sqlstorage import Base
from sqlalchemy.orm import sessionmaker
from zope.sqlalchemy import register as register_session
from .user.model import UserModel, UserCollection
from .group.model import GroupModel, GroupCollection



import copy
import importlib
import os
import threading
import typing
import warnings
from importlib import import_module
from urllib.parse import urlparse

import morepath
import pytz
import sqlalchemy.orm
import transaction
import yaml
from cryptography.fernet import Fernet
from elasticsearch import Elasticsearch
from morepath.publish import resolve_model
from morepath.request import SAME_APP, LinkError
from morepath.request import Request as BaseRequest
from morepath.traject import parse_path
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.pool import NullPool, QueuePool
from zope.sqlalchemy import ZopeTransactionEvents
from zope.sqlalchemy import register as register_session

import morpfw

from .exc import ConfigurationError

threadlocal = threading.local()

default_settings = open(
    os.path.join(os.path.dirname(__file__), "default_settings.yml")
).read()
default_settings = default_settings.replace(r"%(here)s", os.getcwd())
default_settings: dict = yaml.load(default_settings, Loader=yaml.Loader)


class Request(BaseRequest):
    @property
    def session(self):
        return self.environ["beaker.session"]

    @property
    def cache(self):
        return self.environ["beaker.cache"]

    def copy(self, *args, **kwargs):
        """
        Copy the request and environment object.

        This only does a shallow copy, except of wsgi.input
        """
        self.make_body_seekable()
        env = self.environ.copy()
        new_req = self.__class__(env, *args, **kwargs)
        new_req.copy_body()
        new_req.identity = self.identity
        return new_req

    def async_dispatch(self, signal, **kwargs):
        return self.app.async_dispatcher(signal).dispatch(self, **kwargs)

    @property
    def host_url(self):
        """
        The URL through the host (no path)
        """
        e = self.headers
        scheme = e.get("X-FORWARDED-PROTO", None)
        host = e.get("X-FORWARDED-HOST", None)
        port = e.get("X-FORWARDED-PORT", None)
        if scheme and host and not port:
            return "{}://{}".format(scheme, host)
        if scheme and host and port:
            if (scheme == "https" and port == "443") or (
                scheme == "http" and port == "80"
            ):
                return "{}://{}".format(scheme, host)
            return "{}://{}:{}".format(scheme, host, port)
        return super().host_url

    def __enter__(self):
        self._cm_cwd = os.getcwd()
        transaction.begin()
        self.savepoint = transaction.savepoint()

        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if exc_type is not None:
            self.savepoint.rollback()
        self.commit()

    def commit(self):
        transaction.commit()
        if os.path.exists(self._cm_cwd):
            os.chdir(self._cm_cwd)
        self.dispose_db_engines()

    def timezone(self):
        cache_key = "morpfw.cache.timezone"
        self.environ.setdefault(cache_key, None)
        cached_tz = self.environ[cache_key]
        if cached_tz:
            return cached_tz

        user_tz = getattr(self.identity, "timezone", None)
        if user_tz:
            self.environ[cache_key] = user_tz()
        else:
            self.environ[cache_key] = pytz.UTC
        return self.environ[cache_key]

    def permits(self, model, permission):
        if isinstance(model, str):
            model = self.resolve_path(model)
        if isinstance(permission, str):
            perm_mod, perm_cls = permission.split(":")
            mod = import_module(perm_mod)
            klass = getattr(mod, perm_cls)
        else:
            klass = permission
        return self.app._permits(self.identity, model, klass)

    def get_collection(self, type_name):
        self.environ.setdefault("morpfw.cache.collection", {})
        cachemgr = self.environ["morpfw.cache.collection"]
        if type_name in cachemgr:
            return cachemgr[type_name]

        typeinfo = self.app.config.type_registry.get_typeinfo(
            name=type_name, request=self
        )
        col = typeinfo["collection_factory"](self)

        cachemgr[type_name] = col
        return col

    def get_typeinfo(self, type_name):
        typeinfo = self.app.config.type_registry.get_typeinfo(
            name=type_name, request=self
        )
        return typeinfo

    def resolve_path(self, path=None, app=SAME_APP):

        if path is None:
            path = self.path

        if app is None:
            raise LinkError("Cannot path: app is None")

        if app is SAME_APP:
            app = self.app

        self.environ.setdefault("morpfw.cache.path_resolver", {})
        cachemgr = self.environ["morpfw.cache.path_resolver"]
        cache_key = ",".join([str(hash(app)), path])
        if cache_key in cachemgr:
            return cachemgr[cache_key]

        request = self.__class__(self.environ.copy(), app, path_info=path)
        # try to resolve imports..
        result = resolve_model(request)
        cachemgr[cache_key] = result
        return result

    def fernet_encrypt(self, text, encoding="utf8"):
        key = self.app.get_config("morpfw.secret.fernet_key")
        fernet = Fernet(key)
        return fernet.encrypt(text.encode(encoding))

    def fernet_decrypt(self, token, ttl=None, encoding="utf8"):
        token = token.encode(encoding)
        key = self.app.get_config("morpfw.secret.fernet_key")
        fernet = Fernet(key)
        return fernet.decrypt(token, ttl=ttl).decode(encoding)

    def metalink(self, obj, view_name=None):
        return self.app.get_metalink(obj, self, view_name=view_name)

    def resolve_metalink(self, link):
        return self.app.resolve_metalink(link, self)


class DBSessionRequest(Request):
    @property
    def _db_session(self):
        sessions = getattr(threadlocal, "mfw_db_sessions", None)
        if sessions is None:
            sessions = {}
            threadlocal.mfw_db_sessions = sessions

        return threadlocal.mfw_db_sessions

    @property
    def _db_engines(self):
        engines = getattr(threadlocal, "mfw_db_engines", None)
        if engines is None:
            engines = {}
            threadlocal.mfw_db_engines = engines

        return threadlocal.mfw_db_engines

    @property
    def db_session(self) -> sqlalchemy.orm.Session:
        return self.get_db_session("default")

    def get_db_engine(self, name="default"):
        # FIXME: Allow safe pooling, especially on the
        # worker side. Currently NullPool is used
        # to force worker to clean up connections
        # upon completion of tasks
        existing = self._db_engines.get(name, None)
        if existing:
            return existing

        settings = self.app._raw_settings
        config = settings["configuration"]

        home_env = self.app.home_env
        if home_env not in os.environ:
            if "MORP_WORKDIR" in os.environ:
                warnings.warn(
                    "MORP_WORKDIR environment is deprecated, use %s" % home_env,
                    DeprecationWarning,
                )
                home_env = "MORP_WORKDIR"

        cwd = os.environ.get(home_env, os.getcwd())
        os.chdir(cwd)

        key = "morpfw.storage.sqlstorage.dburl"
        if "://" in name:
            dburl = name
        else:
            if name != "default":
                key = "morpfw.storage.sqlstorage.dburl.{}".format(name)

            if not config.get(key, None):
                raise ConfigurationError("{} not found".format(key))

            dburl = config[key]
        engine = sqlalchemy.create_engine(
            dburl, poolclass=NullPool, connect_args={"options": "-c timezone=utc"}
        )

        self._db_engines[name] = engine
        return engine

    def get_db_session(self, name="default"):

        if self._db_session.get(name, None) is None:
            engine = self.get_db_engine(name)
            Session = sessionmaker(bind=engine)
            register_session(Session)
            self._db_session[name] = scoped_session(Session)
        return self._db_session[name]

    def clear_db_session(self, name=None):
        if name:
            if name in self._db_session.keys():
                self._db_session[name].expunge_all()
                self._db_session[name].close()
                del self._db_session[name]
        else:
            for k in list(self._db_session.keys()):
                self._db_session[k].expunge_all()
                self._db_session[k].close()
                self._db_session[k] = None
                del self._db_session[k]
        self.environ["morpfw.memoize"] = {}  # noqa

    def expunge_all(self, name=None):
        if name:
            if name in self._db_session.keys():
                self._db_session[name].expunge_all()
        else:
            for k in list(self._db_session.keys()):
                self._db_session[k].expunge_all()
        self.environ["morpfw.memoize"] = {}  # noqa

    def dispose_db_engines(self, name=None):
        if name:
            if name in self._db_engines.keys():
                if name in self._db_session.keys():
                    self._db_session[name].expunge_all()
                    self._db_session[name].close()
                    del self._db_session[name]
                self._db_engines[name].dispose()
                del self._db_engines[name]
        else:
            for k in list(self._db_engines.keys()):
                if k in self._db_session.keys():
                    self._db_session[k].expunge_all()
                    self._db_session[k].close()
                    del self._db_session[k]
                self._db_engines[k].dispose()
                del self._db_engines[k]
        self.environ["morpfw.memoize"] = {}  # noqa


class ESCapableRequest(DBSessionRequest):
    @property
    def _es_client(self):
        clients = getattr(threadlocal, "mfw_es_clients", None)
        if clients is None:
            clients = {}
            threadlocal.mfw_es_clients = clients

        return threadlocal.mfw_es_clients

    def get_es_client(self, name="default"):

        settings = self.app._raw_settings
        config = settings["configuration"]

        if self._es_client.get(name, None) is None:
            if name == "default":
                key = "morpfw.storage.elasticsearch.hosts"
            else:
                key = "morpfw.storage.elasticsearch.hosts.%s" % name
            hosts = config.get(key, None)
            if not hosts:
                raise ConfigurationError("%s not configured" % key)
            client = Elasticsearch(hosts=hosts)
            self._es_client[name] = client

        return self._es_client[name]


COMMITTED_APPS = []


def request_factory(
    settings: dict,
    extra_environ: typing.Optional[dict] = None,
    scan: bool = True,
    app_factory_opts: typing.Optional[dict] = None,
    path: str = "/",
    request_method: str = "GET",
):
    app_factory_opts = app_factory_opts or {}
    app_factory_opts["scan"] = scan
    extra_environ = extra_environ or {}

    s = copy.deepcopy(default_settings)
    for k in settings.keys():
        if k in s.keys():
            for j, v in settings[k].items():
                s[k][j] = v
        else:
            s[k] = settings[k]

    settings = s

    if "application" not in settings:
        raise KeyError("'application' section is required in settings")
    if "factory" not in settings["application"]:
        raise KeyError("Missing application:factory entry in settings")
    if "class" not in settings["application"]:
        raise KeyError("Missing application:class entry in settings")

    factory_path = settings["application"]["factory"]
    mod, fname = factory_path.split(":")
    factory = getattr(importlib.import_module(mod), fname)

    app_path = settings["application"]["class"]
    mod, clsname = app_path.split(":")
    app_cls = getattr(importlib.import_module(mod), clsname)

    server_url = settings.get("server", {}).get("server_url", "http://localhost")
    parsed = urlparse(server_url)
    if "?" not in path:
        path += "?"
    path_info, qs = path.split("?")
    environ = {
        "PATH_INFO": path,
        "QUERY_STRING": qs,
        "wsgi.url_scheme": parsed.scheme,
        "SERVER_PROTOCOL": "HTTP/1.1",
        "HTTP_HOST": parsed.netloc,
        "REQUEST_METHOD": request_method,
    }
    environ.update(extra_environ)

    sys_environ = settings.get("environment", {}) or {}
    for k, v in sys_environ.items():
        if k not in os.environ.keys():
            os.environ[k] = v

    app = factory(settings, **app_factory_opts)
    wrapped_app = app
    app(environ, lambda *args: (lambda chunk: None))
    while not isinstance(app, morepath.App):
        wrapped = getattr(app, "app", None)
        if wrapped:
            app = wrapped
        else:
            raise ValueError("Unable to locate app object from middleware")
    environ["morpfw.wsgi.app"] = wrapped_app
    return app.request_class(app=app, environ=environ)



import os

import yaml
from more.jwtauth import JWTIdentityPolicy
from more.transaction import TransactionApp
from morepath.reify import reify
from morepath.request import Request
from morpfw.app import SQLApp
from morpfw.authn.pas.app import App
from morpfw.authn.pas.path import hook_auth_models
from morpfw.authz.pas import DefaultAuthzPolicy
from morpfw.sql import Base

from ..common import create_admin, get_client
from .test_auth import _test_authentication


class SQLStorageApp(SQLApp, DefaultAuthzPolicy):
    pass


SQLStorageApp.hook_auth_models()
SQLStorageApp.hook_oauth_models()


def test_authentication_sqlstorage(pgsql_db):
    config = os.path.join(os.path.dirname(__file__), "settings-sqlalchemy.yml")

    c = get_client(config)
    req = c.mfw_request
    Base.metadata.create_all(bind=req.db_session.bind)
    create_admin(c.mfw_request, "admin", "password", "admin@localhost.localdomain")
    _test_authentication(c)



import jsl
from morepath.request import Request
from rulez import compile_condition

from .base import BaseStorage

DATA = {}


class MemoryStorage(BaseStorage):

    incremental_id = False
    incremental_column = "id"

    @property
    def datastore(self):
        return DATA[self.typekey]

    @property
    def typekey(self):
        return ":".join([self.__module__, self.__class__.__name__])

    def __init__(self, request, blobstorage=None):
        DATA.setdefault(self.typekey, {})
        super(MemoryStorage, self).__init__(request, blobstorage)

    def create(self, collection, data):
        assert data["uuid"] is not None
        data = data.copy()
        obj = self.model(self.request, collection, data)
        if self.incremental_id:
            obj.data[self.incremental_column] = len(DATA[self.typekey]) + 1
        identifier = obj.identifier
        DATA[self.typekey][identifier] = obj
        return obj

    def aggregate(self, query=None, group=None, order_by=None, limit=None):
        items = self.search(query)

        result = []
        if group:
            row = {}
            for k, v in group.items():
                if v["function"] == "count":
                    row[k] = len(DATA[self.typekey].values())
                else:
                    raise NotImplementedError
            result.append(row)
        return result

    def search(self, collection, query=None, offset=None, limit=None, order_by=None):
        res = []
        if query:
            f = compile_condition("native", query)
            for o in DATA[self.typekey].values():
                if f(o.data):
                    res.append(o)
        else:
            res = list(DATA[self.typekey].values())
        for r in res:
            r.request = self.request
        if offset is not None:
            res = res[offset:]
        if limit is not None:
            res = res[:limit]
        if order_by is not None:
            col, d = order_by
            res = list(sorted(res, key=lambda x: x.data[col]))
            if d == "desc":
                res = list(reversed(res))
        return res

    def get(self, collection, identifier):
        if identifier not in DATA[self.typekey].keys():
            return None
        res = DATA[self.typekey][identifier]
        res.request = self.request
        return res

    def get_by_id(self, collection, id):
        id_field = self.incremental_column
        data_by_id = {}
        for u, v in DATA[self.typekey].items():
            if id_field not in v.data.keys():
                raise AttributeError("%s does not have %s field" % (v, id_field))
            data_by_id[v.data[id_field]] = v

        if id not in data_by_id.keys():
            return None
        res = data_by_id[id]
        res.request = self.request
        return res

    def get_by_uuid(self, collection, uuid):
        uuid_field = self.app.get_uuidfield(self.model.schema)
        data_by_uuid = {}
        for u, v in DATA[self.typekey].items():
            if uuid_field not in v.data.keys():
                raise AttributeError("%s does not have %s field" % (v, uuid_field))
            data_by_uuid[v.data[uuid_field]] = v

        if uuid not in data_by_uuid.keys():
            return None
        res = data_by_uuid[uuid]
        res.request = self.request
        return res

    def update(self, collection, identifier, data):
        obj = DATA[self.typekey][identifier]
        for k, v in data.items():
            obj.data[k] = v
        return obj

    def delete(self, identifier, model, **kwargs):
        del DATA[self.typekey][identifier]



import morepath

from ..app import App
from .model import GroupCollection, GroupModel, GroupSchema


def get_group(request: morepath.Request, identifier: str) -> GroupModel:
    collection = get_group_collection(request)
    return collection.get_by_groupname(identifier)


def get_group_collection(request: morepath.Request) -> GroupCollection:
    storage = request.app.get_storage(GroupModel, request)
    return GroupCollection(request, storage)



from morpfw import create_app
import morpfw
import yaml
import os
from wsgigzip import gzip
import json
import importlib

APP = {'instance': None}


def wsgi_factory(app_factory):

    @gzip(mime_types=['application/json', 'text/json', 'text/plain', 'text/html'])
    def app(environ, start_response):
        if APP['instance']:
            return APP['instance'](environ, start_response)

        settings = json.loads(os.environ.get('MORP_SETTINGS'))

        application = app_factory(settings)
        APP['instance'] = application
        return application(environ, start_response)

    return app


def app(environ, start_response):
    app_mod, app_fname = os.environ['MORP_APP_FACTORY'].split(':')
    app_factory = getattr(importlib.import_module(app_mod), app_fname)
    return wsgi_factory(app_factory)(environ, start_response)



import morpfw
from morpfw.authz.pas import DefaultAuthzPolicy
from morpfw.crud import permission as crudperm
from morpfw.permission import All


class AppRoot(object):
    def __init__(self, request):
        self.request = request


class App(DefaultAuthzPolicy, morpfw.SQLApp):
    pass


@App.path(model=AppRoot, path="/")
def get_approot(request):
    return AppRoot(request)


@App.permission_rule(model=AppRoot, permission=All)
def allow_all(identity, context, permission):
    """ Default permission rule, allow all """
    return True


@App.json(model=AppRoot)
def index(context, request):
    return {"message": "Hello World"}



from ...app import BaseApp as App
from ...crud import permission as crudperms
from ...crud.model import Collection, Model
from ...permission import All


@App.authz_rule(name="morpfw.fullaccess")
def group_policy(groupname, identity, model, permission):
    request = model.request
    users = request.get_collection("morpfw.pas.user")
    user = users.get_by_userid(identity.userid)
    if user["is_administrator"]:
        return True

    if groupname not in [g["groupname"] for g in user.groups()]:
        return False

    if issubclass(permission, All):
        return True

    return False




import socket

import click

from .cli import cli, load


@cli.command(help="start celery beat scheduler")
@click.pass_context
def scheduler(ctx):
    param = load(ctx.obj["settings"])
    hostname = socket.gethostname()
    ss = param["settings"]["configuration"]["morpfw.celery"]
    # scan
    param["factory"](param["settings"], instantiate=False)
    sched = param["class"].celery.Beat(hostname="scheduler.%s" % hostname, **ss)
    sched.run()




import morepath
from inverter import dc2jsl
from more.jwtauth import ExpiredSignatureError, InvalidTokenError
from morpfw.crud.errors import AlreadyExistsError
from morpfw.crud.validator import validate_schema
from webob.exc import HTTPNotFound

from ....crud import permission as crudperm
from .. import permission
from ..app import App
from ..exc import UserExistsError
from ..path import get_user_collection
from ..policy import verify_refresh_request
from ..utils import has_role, rellink
from ..validator import validate
from .model import (
    CurrentUserModel,
    LoginSchema,
    RegistrationSchema,
    UserCollection,
    UserModel,
    UserSchema,
)


@App.json(
    model=UserCollection,
    name="register",
    request_method="POST",
    permission=permission.Register,
)
def register(context, request):
    """Validate the username and password and create the user."""
    enabled = request.app.get_config("morpfw.new_registration.enabled", True)
    if not enabled:
        raise HTTPNotFound()
    data = request.json
    RegistrationSchema.validate(request, data)

    if data["password"] != data["password_validate"]:

        @request.after
        def adjust_response(response):
            response.status = 422

        return {"status": "error", "message": "Password confirmation does not match"}

    if "state" not in data.keys() or not data["state"]:
        data["state"] = request.app.get_config("morpfw.user.new_user_state")
    del data["password_validate"]
    obj = context.create(data)
    return {"status": "success"}


@App.json(model=UserCollection, name="login")
def login(context, request):
    return {
        "schema": dc2jsl.convert(LoginSchema).get_schema(),
        "links": [rellink(context, request, "login", "POST")],
    }


@App.json(model=UserCollection, name="login", request_method="POST")
def process_login(context, request):
    """Authenticate username and password and log in user"""
    username = request.json["username"]
    password = request.json["password"]

    # Do the password validation.
    user = context.authenticate(username, password)
    if not user:

        @request.after
        def adjust_status(response):
            response.status = 401

        return {
            "status": "error",
            "error": {"code": 401, "message": "Invalid Username / Password"},
        }

    @request.after
    def remember(response):
        """Remember the identity of the user logged in."""
        # We pass the extra info to the identity object.
        response.headers.add("Access-Control-Expose-Headers", "Authorization")
        identity = user.identity
        request.app.remember_identity(response, request, identity)

    return {"status": "success"}


@App.json(model=CurrentUserModel, name="refresh_token")
def refresh_token(context: CurrentUserModel, request: morepath.Request):
    try:
        # Verifies if we're allowed to refresh the token.
        # In this case returns the userid.
        # If not raises exceptions based on InvalidTokenError.
        # If expired this is a ExpiredSignatureError.
        userid = verify_refresh_request(request)
    except ExpiredSignatureError:

        @request.after
        def expired_nonce_or_token(response):
            response.status_code = 403

        return {"status": "error", "message": "Your session has expired."}
    except InvalidTokenError as e:

        @request.after
        def invalid_token(response):
            response.status_code = 403

        return {
            "status": "error",
            "message": "Could not refresh your token. %s" % str(e),
        }
    else:

        @request.after
        def remember(response):
            # create the identity with the userid and updated user info
            identity = context.identity
            # create the updated token and set it in the response header
            request.app.remember_identity(response, request, identity)

        return {"status": "success", "message": "Token successfully refreshed"}


@App.json(model=UserCollection, name="logout")
def logout(context, request):
    """Log out the user."""

    @request.after
    def forget(response):
        request.app.forget_identity(response, request)

    return {"status": "success"}


@App.json(model=UserModel, name="roles", permission=crudperm.View)
def roles(context, request):
    return context.group_roles()


@App.json(
    model=UserModel,
    name="change_password",
    permission=permission.ChangePassword,
    request_method="POST",
)
def admin_change_password(context, request):
    data = request.json
    error = None
    current_password = data.get("password", "")

    if not error and data["new_password"] != data["new_password_validate"]:
        error = "Password confirmation does not match"

    if error:

        @request.after
        def adjust_status(response):
            response.status = 422

        return {"status": "error", "message": error}

    context.change_password(current_password, data["new_password"])
    return {"status": "success"}


@App.json(
    model=CurrentUserModel,
    name="change_password",
    permission=permission.ChangePassword,
    request_method="POST",
)
def user_change_password(context, request):
    data = request.json
    error = None
    current_password = data.get("password", "")

    if not context.validate(current_password, check_state=False):
        error = "Invalid password"

    if not error and data["new_password"] != data["new_password_validate"]:
        error = "Password confirmation does not match"

    if error:

        @request.after
        def adjust_status(response):
            response.status = 422

        return {"status": "error", "message": error}
    context.change_password(current_password, data["new_password"])
    return {"status": "success"}



import os
import typing
from dataclasses import dataclass

import morpfw
import sqlalchemy as sa
from morpfw import sql as morpsql
from morpfw.crud import Collection, Model
from morpfw.crud import permission as crudperm
from morpfw.crud.schema import Schema

from .common import get_client, make_request


class App(morpfw.SQLApp):
    pass


@App.permission_rule(model=Model, permission=crudperm.All)
def allow_all_model_access(identity, context, permission):
    return True


@App.permission_rule(model=Model, permission=crudperm.StateUpdate)
def allow_statemachine(identity, context, permission):
    return True


@App.permission_rule(model=Collection, permission=crudperm.All)
def allow_all_collection_access(identity, context, permission):
    return True


class Page(morpsql.Base):
    __tablename__ = "test_page"

    title = sa.Column(sa.String(length=1024))
    body = sa.Column(sa.Text)


@dataclass
class PageSchema(Schema):
    title: typing.Optional[str] = None
    body: typing.Optional[str] = None


@App.identifierfield(schema=PageSchema)
def page_schema_identifier(schema):
    return "uuid"


class PageCollection(morpfw.Collection):
    schema = PageSchema


class PageModel(morpfw.Model):
    schema = PageSchema


class PageStorage(morpfw.SQLStorage):
    model = PageModel
    orm_model = Page


@App.path(model=PageCollection, path="/")
def get_pagecollection(request):
    storage = PageStorage(request)
    return PageCollection(request, storage)


@App.path(model=PageModel, path="/{identifier}")
def get_page(request, identifier):
    col = get_pagecollection(request)
    return col.get(identifier)


def test_morp_framework(pgsql_db):
    c = get_client(os.path.join(os.path.dirname(__file__), "test_sqlapp-settings.yml"))
    req = make_request(c.app)
    morpsql.Base.metadata.create_all(bind=req.db_session.bind)
    r = c.get("/")

    assert len(r.json["schema"]["properties"]) == 11

    r = c.post_json("/", {"title": "Hello world", "body": "Lorem ipsum"})

    assert r.json["links"][0]["href"].startswith("http://localhost/")
    assert r.json["data"]["title"] == "Hello world"
    assert r.json["data"]["body"] == "Lorem ipsum"

    page_url = r.json["links"][0]["href"]
    r = c.get(page_url)

    assert r.json["data"]["title"] == "Hello world"

    delete_link = r.json["links"][2]
    assert delete_link["method"] == "DELETE"

    r = c.delete(delete_link["href"])

    r = c.get(page_url, expect_errors=True)

    assert r.status_code == 404



import re
from morepath.publish import resolve_model as _resolve_model
from ..interfaces import ISchema
import jsl
import dataclasses
from copy import copy
import typing
from datetime import datetime, date


def resolve_model(request):
    newreq = request.app.request_class(
        request.environ.copy(), request.app, path_info=request.path
    )
    context = _resolve_model(newreq)
    context.request = request
    return context


_marker = object()


def generate_default(schema):
    data = {}
    if isinstance(schema, jsl.DocumentField):
        schema = schema.document_cls
    for n, f in schema._fields.items():
        if isinstance(f, jsl.DocumentField):
            data[n] = generate_default(f)

        else:
            data[n] = f.get_default()
            if data[n] is None:
                if isinstance(f, jsl.StringField):
                    data[n] = None
                elif isinstance(f, jsl.IntField) or isinstance(f, jsl.NumberField):
                    data[n] = None
                elif isinstance(f, jsl.DictField):
                    data[n] = {}
                elif isinstance(f, jsl.ArrayField):
                    data[n] = []

    return data




from urllib.parse import parse_qs, urlencode, urlparse, urlunparse


class CollectionBatching(object):
    def __init__(
        self,
        request,
        collection,
        query=None,
        order_by=None,
        pagesize=20,
        pagenumber=0,
        page_opt="page",
    ):
        self.collection = collection
        self.request = request
        self.pagesize = pagesize
        self.pagenumber = pagenumber
        self.query = query
        self.order_by = order_by
        self._items = None
        self._total = None
        self.page_opt = page_opt

    def items(self):
        if self._items is None:
            offset = self.pagesize * self.pagenumber
            self._items = self.collection.search(
                query=self.query,
                limit=self.pagesize,
                offset=offset,
                order_by=self.order_by,
            )
        return self._items

    def total(self):
        if self._total is None:
            self._total = self.collection.aggregate(
                query=self.query,
                group={"count": {"function": "count", "field": "uuid"}},
            )[0]["count"]
        return self._total

    def total_pages(self):
        total = self.total()
        return int(total / self.pagesize) + 1

    def next_page(self):
        return self.pagenumber + 1

    def navigator(self, url=None, size=3):
        if url is None:
            url = self.request.url
        parsed_url = urlparse(url)
        qs = parse_qs(parsed_url.query)

        def get_page_url(page):
            qso = qs.copy()
            for k, v in list(qso.items()):
                if isinstance(v, list) and len(v) == 1:
                    v = v[0]
                qso[k] = v
            qso[self.page_opt] = page
            return (
                parsed_url.scheme
                + "://"
                + parsed_url.netloc
                + parsed_url.path
                + "?"
                + urlencode(qso)
            )

        pages = []

        minpage = 0
        if (self.pagenumber - size) > 0:
            minpage = self.pagenumber - size
            if self.total_pages() - self.pagenumber <= size:
                minpage -= 1 + size - (self.total_pages() - self.pagenumber)
                if minpage < 0:
                    minpage = 0
            if (self.pagenumber + 1 + size) < self.total_pages():
                minpage += 1

        maxpage = self.total_pages()
        if (self.pagenumber + 1 + size) < self.total_pages():
            maxpage = self.pagenumber + 1 + size
            if self.pagenumber < size:
                maxpage += size - self.pagenumber
                if maxpage > self.total_pages():
                    maxpage = self.total_pages()
            if (self.pagenumber - size) > 0:
                maxpage -= 1

        if self.pagenumber > 0:
            pages.append(
                {
                    "page": self.pagenumber - 1,
                    "title": "<",
                    "url": get_page_url(self.pagenumber - 1),
                }
            )
        elif self.pagenumber == 0:
            pages.append(
                {"page": self.pagenumber, "title": "<", "state": "disabled",}
            )

        if self.pagenumber - size > 0:
            if minpage > 0:
                if self.pagenumber != 0:
                    pages.append({"page": 0, "title": "1", "url": get_page_url(0)})
                pages.append(
                    {"page": None, "title": "...", "url": None, "state": "disabled"}
                )

        # pre
        if self.pagenumber > 0:

            for i in range(minpage, self.pagenumber):
                qs = qs.copy()
                qs["page"] = i
                pages.append({"page": i, "title": str(i + 1), "url": get_page_url(i)})
        pages.append(
            {
                "page": self.pagenumber,
                "title": str(self.pagenumber + 1),
                "url": None,
                "state": "active",
            }
        )

        # post
        if self.pagenumber < self.total_pages():

            for i in range(self.pagenumber + 1, maxpage):
                qs = qs.copy()
                qs["page"] = i
                pages.append({"page": i, "title": str(i + 1), "url": get_page_url(i)})

        if self.pagenumber + 1 + size < self.total_pages():
            if maxpage < self.total_pages():
                pages.append(
                    {"page": None, "title": "...", "url": None, "state": "disabled"}
                )
                if self.pagenumber != self.total_pages() - 1:
                    pages.append(
                        {
                            "page": self.total_pages() - 1,
                            "title": str(self.total_pages()),
                            "url": get_page_url(self.total_pages() - 1),
                        }
                    )
        if self.pagenumber + 1 < self.total_pages():
            pages.append(
                {
                    "page": self.pagenumber + 1,
                    "title": ">",
                    "url": get_page_url(self.pagenumber + 1),
                }
            )
        elif self.pagenumber + 1 == self.total_pages():
            pages.append(
                {
                    "page": self.pagenumber,
                    "title": ">",
                    "url": None,
                    "state": "disabled",
                }
            )

        return pages



import hashlib
import os
from datetime import datetime, timedelta

from pkg_resources import resource_filename

import morepath
import pytz
from webob import static
from webob.exc import HTTPNotFound, HTTPNotModified, HTTPUnauthorized

from .app import BaseApp as App

ETAG = hashlib.md5(datetime.now().strftime(r"%Y%d%d%H").encode("ascii")).hexdigest()


class StaticRoot(object):

    module = "morpfw"
    directory = "static_files"

    def __init__(self, path):
        self.path = path or ""

    def resource_path(self):
        if not self.path.strip():
            return None
        d = resource_filename(self.module, self.directory)
        return os.path.join(d, self.path)


@App.view(model=StaticRoot)
def serve_static(context, request):
    path = context.resource_path()
    if not path:
        raise HTTPNotFound()

    settings = request.app.settings.__dict__
    max_age = int(settings.get("caching", {}).get("max-age", (60 * 60)))

    etag = request.headers.get("If-None-Match", "")

    def add_headers(response):
        response.headers.add("Cache-Control", "public, max-age=%s" % max_age)
        response.headers.add(
            "Expires",
            (datetime.now(tz=pytz.UTC) + timedelta(seconds=max_age)).strftime(
                r"%a, %d %b %Y %H:%M:%S GMT"
            ),
        )
        response.headers.add("ETag", ETAG)

    if etag and etag == ETAG:

        @request.after
        def add_notmodified_headers(response):
            add_headers(response)

        return morepath.Response(status=304)

    resp = request.get_response(static.FileApp(path))
    if resp.status_code == 404:
        raise HTTPNotFound()
    if resp.status_code == 403:
        raise HTTPUnauthorized()

    @request.after
    def add_caching_headers(response):
        add_headers(response)

    return resp




from .app import App
from .exc import AuthManagerError


@App.json(model=AuthManagerError)
def exception_view(context, request):
    @request.after
    def adjust_status(response):
        response.status = 422
    return {
        'status': 'error',
        'type': context.__class__.__name__,
        'message': '%s ( %s )' % (
            context.__class__.__name__,
            str(context))
    }



import copy
import dataclasses
import json
import logging
import re
import warnings
from logging import warn
from uuid import uuid4

import morepath
import rulez
from DateTime import DateTime
from inverter import dc2colanderjson, dc2jsl
from jsonschema import ValidationError as JSLValidationError
from jsonschema import validate
from jsonschema.validators import Draft4Validator
from morepath import reify
from rulez import OperatorNotAllowedError
from rulez import field as rfield
from rulez import parse_dsl, validate_condition
from transitions import Machine

from ..interfaces import ICollection, IModel, IStorage
from ..memoizer import requestmemoize
from ..request import Request
from . import permission, signals
from .const import SEPARATOR
from .errors import (
    AlreadyExistsError,
    BlobStorageNotImplementedError,
    FormValidationError,
    StateUpdateProhibitedError,
    UnprocessableError,
    ValidationError,
)
from .log import logger
from .relationship import BackReferenceResolver, ReferenceResolver

ALLOWED_SEARCH_OPERATORS = [
    "and",
    "or",
    "==",
    "in",
    "~",
    "!=",
    ">",
    "<",
    ">=",
    "<=",
    "match",
]

_marker = object()


class Collection(ICollection):

    create_view_enabled = True
    search_view_enabled = True
    search_allow_queryobject = True
    aggregate_view_enabled = True

    exist_exc = AlreadyExistsError

    @property
    def schema(self):
        raise NotImplementedError

    @property
    def name(self):
        typeinfo = self.request.app.get_typeinfo_by_schema(
            request=self.request, schema=self.schema
        )
        return typeinfo["name"]

    def __setitem__(self, key, value):
        if key not in self.schema.__dataclass_fields__:
            raise KeyError(key)
        self.data[key] = value

    def __getitem__(self, key):
        if key not in self.schema.__dataclass_fields__:
            raise KeyError(key)
        return self.data[key]

    def __delitem__(self, key):
        if key not in self.schema.__dataclass_fields__:
            raise KeyError(key)
        del self.data[key]

    def __init__(self, request: Request, storage: IStorage, data=None):
        self.request = request
        self.app = request.app
        self.storage = storage
        self.data = None
        if data:
            # FIXME: what is this for again? o_O
            self.data = request.app.get_dataprovider(self.schema, data, self.storage)

    def search(self, query=None, offset=0, limit=None, order_by=None, secure=False):
        objs = self._search(query, offset, limit, order_by, secure)
        if secure and limit:
            nextpage = {
                "query": query,
                "offset": offset + limit,
                "limit": limit,
                "order_by": order_by,
            }
            while len(objs) < limit:
                nextobjs = self._search(secure=True, **nextpage)
                if len(nextobjs) == 0:
                    return list(objs[:limit])
                nextpage["offset"] = nextpage["offset"] + limit
                objs = objs + nextobjs
        return objs

    @requestmemoize()
    def all(self):
        return self.search()

    @requestmemoize()
    def count(self):
        return self.aggregate(group={"count": {"function": "count", "field": "uuid"}})[
            0
        ]["count"]

    @requestmemoize()
    def max_id(self):
        return self.aggregate(group={"max_id": {"function": "max", "field": "id"}})[0][
            "max_id"
        ]

    @requestmemoize()
    def min_id(self):
        return self.aggregate(group={"min_id": {"function": "min", "field": "id"}})[0][
            "min_id"
        ]

    def _search(self, query=None, offset=0, limit=None, order_by=None, secure=False):
        if query:
            validate_condition(query, ALLOWED_SEARCH_OPERATORS)

        if order_by is None:
            order_by = ("created", "desc")

        objs = self.storage.search(
            self, query, offset=offset, limit=limit, order_by=order_by
        )

        if secure:
            objs = list(
                [
                    obj
                    for obj in objs
                    if self.request.app.permits(self.request, obj, permission.View)
                ]
            )
        return list(objs)

    def aggregate(self, query=None, group=None, order_by=None, limit=None):
        if query:
            validate_condition(query, ALLOWED_SEARCH_OPERATORS)
        prov = self.aggregateprovider()
        objs = prov.aggregate(query, group=group, order_by=order_by, limit=limit)
        return list(objs)

    def searchprovider(self):
        if self.app.get_searchprovider.by_args(self).all_matches:
            return self.app.get_searchprovider(self)
        return None

    def aggregateprovider(self):
        if self.app.get_aggregateprovider.by_args(self).all_matches:
            return self.app.get_aggregateprovider(self)
        return None

    def create(self, data, deserialize=True, secure=False):
        if secure:
            if "state" in data:
                raise StateUpdateProhibitedError()
            for protected in self.schema.__protected_fields__:
                if protected in data.keys():
                    raise UnprocessableError(
                        "%s is not allowed to be set in this context" % protected
                    )
            for fn, fo in self.schema.__dataclass_fields__.items():
                if fn in data.keys() and not fo.metadata.get("initializable", True):
                    raise UnprocessableError(
                        "%s is not allowed to be set in this context" % fn
                    )
        data = self.schema.validate(
            self.request, data, deserialize=deserialize, context=self
        )
        self.before_create(data)
        identifier = self.app.get_default_identifier(self.schema, data, self.request)
        if identifier and self.get(identifier):
            raise self.exist_exc(identifier)
        data = self.storage.set_schema_defaults(data)
        for fname, field in self.schema.__dataclass_fields__.items():
            if data[fname] is not None:
                continue
            dc_default_factory = field.default_factory
            if not isinstance(dc_default_factory, dataclasses._MISSING_TYPE):
                data[fname] = dc_default_factory()

            if data[fname] is not None:
                continue

            default_factory = field.metadata.get("default_factory", None)
            if default_factory:
                data[fname] = default_factory(self, self.request)
        unique_constraint = getattr(self.schema, "__unique_constraint__", None)
        if unique_constraint:
            unique_search = []
            msg = []
            for c in unique_constraint:
                unique_search.append(rulez.field[c] == data[c])
                msg.append(f"{c}=({data[c]})")
            if self.search(rulez.and_(*unique_search)):
                raise self.exist_exc(" ".join(msg))

        self.update_computed_fields(data)
        obj = self._create(data)
        obj.set_initial_state()
        dispatch = self.request.app.dispatcher(signals.OBJECT_CREATED)
        dispatch.dispatch(self.request, obj)
        obj.after_created()
        obj.save()
        return obj

    def update_computed_fields(self, data):
        for fn, field in self.schema.__dataclass_fields__.items():
            compute = field.metadata.get("compute_value", None)
            if compute:
                data[fn] = compute(self.request, data, None)

    def _create(self, data):
        return self.storage.create(self, data)

    @requestmemoize()
    def get(self, identifier):
        if isinstance(identifier, list) or isinstance(identifier, tuple):
            identifier = self.request.app.join_identifier(*identifier)
        return self.storage.get(self, identifier)

    @requestmemoize()
    def get_by_uuid(self, uuid):
        return self.storage.get_by_uuid(self, uuid)

    def json(self):
        return {
            "schema": dc2jsl.convert(self.schema).get_schema(ordered=True),
            "links": self.links(),
        }

    def links(self):
        request = self.request
        links = []
        if self.create_view_enabled:
            links.append(
                {"rel": "create", "href": request.link(self), "method": "POST"}
            )
        if self.search_view_enabled:
            links.append({"rel": "search", "href": request.link(self, "+search")})
        if self.aggregate_view_enabled:
            links.append({"rel": "aggregate", "href": request.link(self, "+aggregate")})
        links += self._links()
        return links

    def _links(self):
        return []


class Model(IModel):

    linkable = True
    update_view_enabled = True
    delete_view_enabled = True

    blobstorage_field: str = "blobs"
    blob_fields: list = []
    blob_field_options: dict = {}

    @property
    def hidden_fields(self) -> list:
        hidden = set()
        for fn, fo in self.schema.__dataclass_fields__.items():
            if fo.metadata.get("hidden", False):
                hidden.add(fn)
        return hidden

    def __setitem__(self, key, value):
        warnings.warn(
            "Value assignment through model will be removed", DeprecationWarning
        )
        self.data[key] = value

    def __getitem__(self, key):
        return self.data[key]

    def __delitem__(self, key):
        warnings.warn(
            "Value deletion through model will be removed", DeprecationWarning
        )
        del self.data[key]

    def __dict__(self):
        return self.data.as_dict()

    def title(self):
        fields = list(self.schema.__dataclass_fields__.keys())
        if "title" in fields:
            if self["title"]:
                return self["title"]
        if "name" in fields:
            if self["name"]:
                return self["name"]
        return "%s:%s" % (str(self.__class__.__name__), self["uuid"])

    @property
    def statemachine_view_enabled(self):
        if self.statemachine():
            return True
        return False

    @property
    def xattr_view_enabled(self):
        if self.xattrprovider():
            return True
        return False

    @property
    def schema(self):
        raise NotImplementedError

    @property
    def identifier(self):
        if self._cached_identifier:
            return self._cached_identifier
        idfield = self.app.get_identifierfield(self.schema)
        identifier = self.data.get(idfield)
        if identifier is None:
            identifier = self.app.get_default_identifier(
                self.schema, self.data, self.request
            )
            if identifier is None:
                return None
            return identifier
        self._cached_identifier = identifier
        return identifier

    @property
    def uuid(self):
        uuid_field = self.app.get_uuidfield(self.schema)
        return self.data[uuid_field]

    def __init__(self, request, collection, data):
        self.collection = collection
        self.request = request
        self.storage = collection.storage
        self.app = request.app
        self.data = request.app.get_dataprovider(self.schema, data, self.storage)
        self._cached_identifier = None
        super().__init__(request, collection, data)

    def is_editable(self):
        sm = self.statemachine()
        if sm:
            if sm.is_readonly():
                return False
        return True

    def update(self, newdata: dict, secure: bool = False, deserialize: bool = True):
        if secure:
            if "state" in newdata:
                raise StateUpdateProhibitedError()
            for protected in self.schema.__protected_fields__:
                if protected in newdata.keys():
                    raise UnprocessableError(
                        "%s is not allowed to be updated in this context" % protected
                    )
            for fn, fo in self.schema.__dataclass_fields__.items():
                if fn in newdata.keys() and not fo.metadata.get("editable", True):
                    raise UnprocessableError(
                        "%s is not allowed to be updated in this context" % fn
                    )
            if not self.is_editable():
                raise UnprocessableError(
                    "Record is not allowed to be updated in state %s" % sm.state
                )

        if deserialize:
            data = self.data.as_json()
        else:
            data = self.data.as_dict()
        self.before_update(newdata)
        data.update(newdata)
        self.update_computed_fields(data)
        self.schema.validate(
            self.request, data, deserialize=deserialize, update_mode=True, context=self
        )
        unique_constraint = getattr(self.schema, "__unique_constraint__", None)
        if unique_constraint:
            unique_search = []
            msg = []
            for c in unique_constraint:
                unique_search.append(rulez.field[c] == data[c])
                msg.append(f"{c}=({data[c]})")
            res = self.collection.search(rulez.and_(*unique_search))
            if res:
                if res[0].identifier != self.identifier:
                    raise self.collection.exist_exc(" ".join(msg))

        if deserialize:
            cschema = dc2colanderjson.convert(self.schema, request=self.request)
            cs = cschema()
            cs = cs.bind(context=self, request=self.request)
            data = cs.deserialize(data)
        self.storage.update(self.collection, self.identifier, data)
        dispatch = self.request.app.dispatcher(signals.OBJECT_UPDATED)
        dispatch.dispatch(self.request, self)
        self.after_updated()

    def update_computed_fields(self, data):
        for fn, field in self.schema.__dataclass_fields__.items():
            compute = field.metadata.get("compute_value", None)
            if compute:
                data[fn] = compute(self.request, data, self)

    def delete(self, *, cascade=True, **kwargs):
        dispatch = self.request.app.dispatcher(signals.OBJECT_TOBEDELETED)
        dispatch.dispatch(self.request, self)

        if cascade:
            for bref in self.backreferences().values():
                for refitem in self.resolve_backreference(bref):
                    refitem.delete(cascade=cascade)

        if not self.before_delete():
            return
        blob_uuids = []
        for blobfield in self.blob_fields:
            if self.blobstorage_field not in self.data.keys():
                uuid = None
            elif not self.data[self.blobstorage_field]:
                uuid = None
            elif not blobfield in self.data[self.blobstorage_field]:
                uuid = None
            else:
                uuid = self.data[self.blobstorage_field][blobfield]
            if uuid:
                blob_uuids.append(uuid)
        self.storage.delete(self.identifier, model=self, **kwargs)
        for blob_uuid in blob_uuids:
            self.storage.delete_blob(blob_uuid)

    def save(self):
        if self.data.changed:
            data = self.as_dict()
            data = self.schema.validate(
                self.request, data, deserialize=False, context=self
            )
            self.storage.update(self.collection, self.identifier, data)

    def _base_json(self, exclude_metadata=False):

        exclude_fields = self.hidden_fields
        if exclude_metadata:
            from .schema import Schema

            exclude_fields += list(Schema.__dataclass_fields__.keys())
        cschema = dc2colanderjson.convert(
            self.schema, exclude_fields=exclude_fields, request=self.request
        )
        cs = cschema()
        cs = cs.bind(context=self, request=self.request)
        return cs.serialize(self.data.as_dict())

    @requestmemoize()
    def base_json(self):
        return self._base_json()

    @requestmemoize()
    def data_json(self):
        return self._base_json(exclude_metadata=True)

    def _json(self):
        return self.base_json()

    def json(self):
        if self.linkable:
            return {"data": self._json(), "links": self.links()}
        return self._json()

    def as_json(self):
        return self.data.as_json()

    def as_dict(self):
        return self.data.as_dict()

    @requestmemoize()
    def links(self):
        links = []
        links.append({"rel": "self", "href": self.request.link(self)})
        if self.update_view_enabled:
            links.append(
                {"rel": "update", "href": self.request.link(self), "method": "PATCH"}
            )
        if self.delete_view_enabled:
            links.append(
                {"rel": "delete", "href": self.request.link(self), "method": "DELETE"}
            )
        if self.statemachine_view_enabled:
            links.append(
                {
                    "rel": "statemachine",
                    "href": self.request.link(self, "+statemachine"),
                    "method": "POST",
                }
            )
        links += self._links()
        return links

    def _links(self):
        return []

    @requestmemoize()
    def rulesprovider(self):
        return self.app.get_rulesprovider(self)

    @requestmemoize()
    def statemachine(self):
        if self.app.get_statemachine.by_args(self).all_matches:
            return self.app.get_statemachine(self)
        return None

    @requestmemoize()
    def state_title(self):
        sm = self.statemachine()
        if sm:
            return sm.state.replace("_", " ").title()
        return None

    @requestmemoize()
    def state(self):
        sm = self.statemachine()
        if sm:
            return sm.state
        return None

    @requestmemoize()
    def xattrprovider(self):
        if self.app.get_xattrprovider.by_args(self).all_matches:
            return self.app.get_xattrprovider(self)
        return None

    def set_initial_state(self):
        self.statemachine()

    def _blob_guard(self, field):
        if self.blobstorage_field not in self.schema.__dataclass_fields__.keys():
            raise BlobStorageNotImplementedError(
                "Object does not implement blobs store"
            )
        if field not in self.blob_fields:
            raise BlobStorageNotImplementedError(
                "Field %s not allowed for blobstorage" % field
            )

    def put_blob(
        self, field, fileobj, filename, mimetype=None, size=None, encoding=None
    ):
        self._blob_guard(field)
        allowed_types = self.blob_field_options.get(field, {}).get("allowed_types", [])
        if mimetype and allowed_types and mimetype not in allowed_types:
            raise ValueError("Mimetype %s not allowed" % mimetype)
        self.before_blobput(field, fileobj, filename, mimetype, size, encoding)
        blob_data = self.data[self.blobstorage_field] or {}
        existing = blob_data.get(field, None)
        blob = self.storage.put_blob(fileobj, filename, mimetype, size, encoding)
        blob_data[field] = blob.uuid
        self.update({self.blobstorage_field: blob_data})
        if existing:
            self.storage.delete_blob(existing)
        self.after_blobput(field, blob)
        self.save()
        return blob

    def get_blob(self, field):
        if self.blobstorage_field not in self.data.keys():
            return None
        if not self.data[self.blobstorage_field]:
            return None
        if not field in self.data[self.blobstorage_field]:
            return None
        uuid = self.data[self.blobstorage_field][field]
        blob = self.storage.get_blob(uuid)
        return blob

    def delete_blob(self, field):
        if self.blobstorage_field not in self.data.keys():
            return None
        if not self.data[self.blobstorage_field]:
            return None
        if not field in self.data[self.blobstorage_field]:
            return None
        uuid = self.data[self.blobstorage_field][field]
        if not self.before_blobdelete(field):
            return
        self.storage.delete_blob(uuid)
        self.save()

    def references(self):
        result = {}
        refs = getattr(self.schema, "__references__", None)
        if refs:
            for ref in refs:
                result[ref.name] = ref
        return result

    def backreferences(self):
        result = {}
        refs = getattr(self.schema, "__backreferences__", None)
        if refs:
            for ref in refs:
                result[ref.name] = ref
        return result

    def resolve_reference(self, reference):
        resolver = ReferenceResolver(self.request, self, reference)
        return resolver.resolve()

    def resolve_backreference(self, backreference):
        resolver = BackReferenceResolver(self.request, self, backreference)
        return resolver.resolve()




import os
import time
import typing
import uuid
from dataclasses import dataclass

import jsl
import morepath
import morpfw
import pytest
import sqlalchemy as sa
from morpfw import sql as morpsql
from morpfw.crud import Collection, Model
from morpfw.crud import permission as crudperm
from morpfw.crud.schema import Schema

from .common import get_client, make_request, start_scheduler, start_worker
from .test_sqlapp import Page


class App(morpfw.SQLApp):
    pass


@App.permission_rule(model=Model, permission=crudperm.All)
def allow_all_model_access(identity, context, permission):
    return True


@App.permission_rule(model=Collection, permission=crudperm.All)
def allow_all_collection_access(identity, context, permission):
    return True


@dataclass
class PageSchema(Schema):
    title: typing.Optional[str] = None
    body: typing.Optional[str] = None


@App.identifierfield(schema=PageSchema)
def page_schema_identifier(schema):
    return "uuid"


class PageCollection(morpfw.Collection):
    schema = PageSchema


class PageModel(morpfw.Model):
    schema = PageSchema


class PageStorage(morpfw.SQLStorage):
    model = PageModel
    orm_model = Page


@App.path(model=PageCollection, path="/")
def get_pagecollection(request):
    storage = PageStorage(request)
    return PageCollection(request, storage)


@App.path(model=PageModel, path="/{identifier}")
def get_page(request, identifier):
    storage = PageStorage(request)
    collection = PageCollection(request, storage)
    return storage.get(collection, identifier)


@App.periodic(name="test-tick", seconds=1)
def tick(request_options):
    request_options["scan"] = False
    morepath.scan(morpfw)
    with morpfw.request_factory(**request_options) as request:
        collection = get_pagecollection(request)
        collection.create({"title": "Hello", "body": "World"})
    print("tick")


def test_scheduler(pgsql_db, pika_connection_channel):
    config = os.path.join(
        os.path.dirname(__file__), "test_workerscheduler-settings.yml"
    )
    c = get_client(config)
    request = make_request(c.app)
    morpsql.Base.metadata.create_all(bind=request.db_session.bind)

    r = c.get("/+search")

    assert len(r.json["results"]) == 0

    worker_proc = start_worker(c.app)
    time.sleep(2)
    sched_proc = start_scheduler(c.app)
    time.sleep(2)

    r = c.get("/+search")

    assert len(r.json["results"]) > 0

    sched_proc.terminate()
    time.sleep(2)
    worker_proc.terminate()
    time.sleep(5)



import typing
from dataclasses import dataclass
import morpfw
from .app import App
from .model import PageModel


class PageStateMachine(morpfw.StateMachine):

    states = ['new', 'pending', 'approved']
    transitions = [
        {'trigger': 'approve', 'source': [
            'new', 'pending'], 'dest': 'approved'},
        {'trigger': 'submit', 'source': 'new', 'dest': 'pending'}
    ]


@App.statemachine(model=PageModel)
def get_pagemodel_statemachine(context):
    return PageStateMachine(context)



import tempfile
import typing
from dataclasses import dataclass, field
from datetime import date
from urllib.parse import quote

import rulez
from morpfw import Schema
from morpfw.app import BaseApp
from morpfw.crud import permission as crudperm
from morpfw.crud.model import Collection, Model
from morpfw.crud.statemachine.base import StateMachine
from morpfw.crud.workflow import WorkflowCollection, WorkflowModel


class App(BaseApp):
    pass


@App.permission_rule(model=Model, permission=crudperm.All)
def allow_all_model_access(identity, context, permission):
    return True


@App.permission_rule(model=Model, permission=crudperm.StateUpdate)
def allow_all_model_state_access(identity, context, permission):
    return True


@App.permission_rule(model=Collection, permission=crudperm.All)
def allow_all_collection_access(identity, context, permission):
    return True


def generate_workflow(request, data, model):
    if data["process_uuid"]:
        return data["process_uuid"]
    wfs = request.get_collection("tests.process_workflow")
    wf = wfs.create({})
    return wf.uuid


@dataclass
class OrderSchema(Schema):

    title: str = field(
        default="", metadata={"format": "text"},
    )
    process_uuid: typing.Optional[str] = field(
        default=None, metadata={"format": "uuid", "compute_value": generate_workflow}
    )


class OrderCollection(Collection):
    schema = OrderSchema


class OrderModel(Model):
    schema = OrderSchema


class OrderSM(StateMachine):

    states = ["new", "submitted", "closed", "cancelled"]
    transitions = [
        {"trigger": "submit", "source": "new", "dest": "submitted"},
        {"trigger": "complete", "source": "submitted", "dest": "closed"},
        {"trigger": "cancel", "source": ["new", "submitted"], "dest": "cancelled"},
    ]

    def trigger_workflow(self):
        context = self._context
        request = self._request
        wfs = request.get_collection("tests.process_workflow")
        wf = wfs.get(context["process_uuid"])
        wf.process(context)


@App.statemachine(model=OrderModel)
def order_statemachine(context):
    return OrderSM(context, after_state_change="trigger_workflow")


@App.path(model=OrderCollection, path="orders")
def order_collection_factory(request):
    storage = request.app.get_storage(OrderModel, request)
    return OrderCollection(request, storage)


@App.path(model=OrderModel, path="orders/{identifier}")
def order_model_factory(request, identifier):
    collection = order_collection_factory(request)
    return collection.get(identifier)


@App.typeinfo(name="tests.order", schema=OrderSchema)
def get_order_typeinfo(request):
    return {
        "title": "Order",
        "description": "Order type",
        "schema": OrderSchema,
        "collection": OrderCollection,
        "collection_factory": order_collection_factory,
        "model": OrderModel,
        "model_factory": order_model_factory,
    }


@dataclass
class KitchenOrderSchema(Schema):

    issue: typing.Optional[str] = field(
        default=None, metadata={"format": "text"},
    )
    process_uuid: typing.Optional[str] = field(
        default=None, metadata={"format": "uuid", "compute_value": generate_workflow}
    )


class KitchenOrderCollection(Collection):
    schema = KitchenOrderSchema


class KitchenOrderModel(Model):
    schema = KitchenOrderSchema


class KitchenOrderSM(StateMachine):

    states = ["new", "processing", "completed", "cancelled"]
    transitions = [
        {"trigger": "process", "source": "new", "dest": "processing"},
        {"trigger": "complete", "source": "processing", "dest": "completed"},
        {"trigger": "cancel", "source": ["new", "processing"], "dest": "cancelled"},
    ]

    def trigger_workflow(self):
        context = self._context
        request = self._request
        wfs = request.get_collection("tests.process_workflow")
        if context["process_uuid"] is None:
            wf = wfs.create({})
            context.update({"process_uuid": wf.uuid})
        else:
            wf = wfs.get(context["process_uuid"])
        wf.process(context)


@App.statemachine(model=KitchenOrderModel)
def kitchen_order_statemachine(context):
    return KitchenOrderSM(context, after_state_change="trigger_workflow")


@App.path(model=KitchenOrderCollection, path="kitchen_orders")
def kitchen_order_collection_factory(request):
    storage = request.app.get_storage(KitchenOrderModel, request)
    return KitchenOrderCollection(request, storage)


@App.path(model=KitchenOrderModel, path="kitchen_orders/{identifier}")
def kitchen_order_model_factory(request, identifier):
    collection = kitchen_order_collection_factory(request)
    return collection.get(identifier)


@App.typeinfo(name="tests.kitchen_order", schema=OrderSchema)
def get_kitchen_order_typeinfo(request):
    return {
        "title": "Kitchen Order",
        "description": "Kitchen Order type",
        "schema": KitchenOrderSchema,
        "collection": KitchenOrderCollection,
        "collection_factory": kitchen_order_collection_factory,
        "model": KitchenOrderModel,
        "model_factory": kitchen_order_model_factory,
    }


@dataclass
class DeliveryOrderSchema(Schema):

    receiver: typing.Optional[str] = field(
        default=None, metadata={"format": "text"},
    )
    process_uuid: typing.Optional[str] = field(
        default=None, metadata={"format": "uuid", "compute_value": generate_workflow}
    )


class DeliveryOrderCollection(Collection):
    schema = DeliveryOrderSchema


class DeliveryOrderModel(Model):
    schema = DeliveryOrderSchema


class DeliveryOrderSM(StateMachine):

    states = ["new", "shipped", "delivered", "cancelled"]
    transitions = [
        {"trigger": "ship", "source": "new", "dest": "shipped"},
        {"trigger": "complete", "source": "shipped", "dest": "delivered"},
        {"trigger": "cancel", "source": ["new", "shipped"], "dest": "cancelled"},
    ]

    def trigger_workflow(self):
        context = self._context
        request = self._request
        wfs = request.get_collection("tests.process_workflow")
        wf = wfs.get(context["process_uuid"])
        wf.process(context)


@App.statemachine(model=DeliveryOrderModel)
def delivery_order_statemachine(context):
    return DeliveryOrderSM(context, after_state_change="trigger_workflow")


@App.path(model=DeliveryOrderCollection, path="delivery_orders")
def delivery_order_collection_factory(request):
    storage = request.app.get_storage(DeliveryOrderModel, request)
    return DeliveryOrderCollection(request, storage)


@App.path(model=DeliveryOrderModel, path="delivery_orders/{identifier}")
def delivery_order_model_factory(request, identifier):
    collection = delivery_order_collection_factory(request)
    return collection.get(identifier)


@App.typeinfo(name="tests.delivery_order", schema=OrderSchema)
def get_delivery_order_typeinfo(request):
    return {
        "title": "Delivery Order",
        "description": "Delivery Order type",
        "schema": DeliveryOrderSchema,
        "collection": DeliveryOrderCollection,
        "collection_factory": delivery_order_collection_factory,
        "model": DeliveryOrderModel,
        "model_factory": delivery_order_model_factory,
    }


@dataclass
class ProcessWorkflowSchema(Schema):
    pass


class ProcessWorkflowCollection(WorkflowCollection):
    schema = ProcessWorkflowSchema


class ProcessWorkflowModel(WorkflowModel):
    schema = ProcessWorkflowSchema


@App.path(model=ProcessWorkflowCollection, path="process_workflows")
def process_workflow_collection_factory(request):
    storage = request.app.get_storage(ProcessWorkflowModel, request)
    return ProcessWorkflowCollection(request, storage)


@App.path(model=ProcessWorkflowModel, path="process_workflows/{identifier}")
def process_workflow_model_factory(request, identifier):
    collection = process_workflow_collection_factory(request)
    return collection.get(identifier)


@App.typeinfo(name="tests.process_workflow", schema=OrderSchema)
def get_process_workflow_typeinfo(request):
    return {
        "title": "Process Workflow",
        "description": "Process Workflow type",
        "schema": ProcessWorkflowSchema,
        "collection": ProcessWorkflowCollection,
        "collection_factory": process_workflow_collection_factory,
        "model": ProcessWorkflowModel,
        "model_factory": process_workflow_model_factory,
    }


@ProcessWorkflowModel.transition(model=OrderModel)
def process_order(wf, obj, request):
    kos = request.get_collection("tests.kitchen_order")
    if obj["state"] == "submitted":
        kos.create({"process_uuid": wf.uuid})


@ProcessWorkflowModel.transition(model=KitchenOrderModel)
def process_kitchenorder(wf, obj, request):
    dos = request.get_collection("tests.delivery_order")
    if obj["state"] == "completed":
        dos.create({"process_uuid": wf.uuid})


@ProcessWorkflowModel.transition(model=KitchenOrderModel)
def cancel_kitchenorder(wf, obj, request):
    if obj["state"] == "cancelled":
        # cancel all orders
        orders = request.get_collection("tests.order")
        os = orders.search(rulez.field("process_uuid") == wf.uuid)
        for o in os:
            sm = o.statemachine()
            sm.cancel()


@ProcessWorkflowModel.transition(model=DeliveryOrderModel)
def cancel_deliveryorder(wf, obj, request):
    if obj["state"] == "cancelled":
        # cancel all orders
        orders = request.get_collection("tests.order")
        os = orders.search(rulez.field("process_uuid") == wf.uuid)
        for o in os:
            sm = o.statemachine()
            sm.cancel()


@ProcessWorkflowModel.transition(model=DeliveryOrderModel)
def complete_deliveryorder(wf, obj, request):
    if obj["state"] == "delivered":
        # cancel all orders
        orders = request.get_collection("tests.kitchen_order")
        os = orders.search(rulez.field("process_uuid") == wf.uuid)
        for o in os:
            sm = o.statemachine()
            sm.complete()


def run_test(c):
    c.authorization = ("Basic", ("admin", "admin"))

    r = c.get("/orders")

    r = c.post_json("/orders", {"title": "Test Order"})

    order_uuid = r.json["data"]["uuid"]
    process_uuid = r.json["data"]["process_uuid"]

    r = c.get("/kitchen_orders/+search")
    assert len(r.json["results"]) == 0
    r = c.get("/delivery_orders/+search")
    assert len(r.json["results"]) == 0
    r = c.get("/process_workflows/+search")
    assert len(r.json["results"]) == 1

    r = c.post_json("/orders/%s/+statemachine" % order_uuid, {"transition": "submit"})

    r = c.get("/kitchen_orders/+search")
    assert len(r.json["results"]) == 1
    r = c.get("/process_workflows/+search")
    assert len(r.json["results"]) == 1
    r = c.get("/delivery_orders/+search")
    assert len(r.json["results"]) == 0

    r = c.get(
        "/kitchen_orders/+search?q=%s" % quote('process_uuid == "%s"' % process_uuid)
    )
    kitchenorder_uuid = r.json["results"][0]["data"]["uuid"]

    r = c.post_json(
        "/kitchen_orders/%s/+statemachine" % kitchenorder_uuid,
        {"transition": "process"},
    )

    r = c.get("/kitchen_orders/+search")
    assert len(r.json["results"]) == 1
    r = c.get("/process_workflows/+search")
    assert len(r.json["results"]) == 1
    r = c.get("/delivery_orders/+search")
    assert len(r.json["results"]) == 0

    r = c.post_json(
        "/kitchen_orders/%s/+statemachine" % kitchenorder_uuid,
        {"transition": "complete"},
    )

    r = c.get("/kitchen_orders/+search")
    assert len(r.json["results"]) == 1
    r = c.get("/process_workflows/+search")
    assert len(r.json["results"]) == 1
    r = c.get("/delivery_orders/+search")
    assert len(r.json["results"]) == 1

    r = c.get(
        "/delivery_orders/+search?q=%s" % quote('process_uuid == "%s"' % process_uuid)
    )
    deliveryorder_uuid = r.json["results"][0]["data"]["uuid"]

    r = c.post_json(
        "/delivery_orders/%s/+statemachine" % deliveryorder_uuid,
        {"transition": "cancel"},
    )

    r = c.get("/orders/%s" % order_uuid)
    assert r.json["data"]["state"] == "cancelled"




from .model import GroupCollection, GroupModel
from .schema import GroupSchema
from .path import get_group, get_group_collection
from ..app import App


@App.typeinfo(name='morpfw.pas.group', schema=GroupSchema)
def get_typeinfo(request):
    return {
        'title': 'Group',
        'description': 'Group type',
        'schema': GroupSchema,
        'collection': GroupCollection,
        'collection_factory': get_group_collection,
        'model': GroupModel,
        'model_factory': get_group,
        'internal': True
    }
